from importlib.metadata import version
from pathlib import Path

import zarr
from packaging.version import Version
from pytest import fixture

zarr_version = version("zarr")
ZARR_PYTHON_V = 2 if Version(zarr_version) < Version("3.0.0a") else 3


def _create_zarr(tempdir, zarr_format=2):
    zarr_path = Path(tempdir) / f"test_group_v{2}.zarr"

    if ZARR_PYTHON_V == 3:
        group = zarr.open_group(store=zarr_path, mode="w", zarr_format=zarr_format)
    else:
        group = zarr.open_group(store=zarr_path, mode="w", zarr_version=zarr_format)

    for i in range(3):
        if ZARR_PYTHON_V == 3:
            group.create_array(f"array_{i}", shape=(10, 10), dtype="i4")
        else:
            group.empty(f"array_{i}", shape=(10, 10), dtype="i4")

    for i in range(3):
        group.create_group(f"group_{i}")

    return zarr_path


@fixture
def local_zarr_path_v2(tmpdir) -> tuple[Path, int]:
    zarr_path = _create_zarr(tmpdir, zarr_format=2)
    return zarr_path, 2


@fixture
def local_zarr_path_v3(tmpdir) -> tuple[Path, int]:
    zarr_path = _create_zarr(tmpdir, zarr_format=3)
    return zarr_path, 3


@fixture
def local_zarr_str_v2(tmpdir) -> tuple[str, int]:
    zarr_path = _create_zarr(tmpdir, zarr_format=2)
    return str(zarr_path.absolute()), 2


@fixture
def local_zarr_str_v3(tmpdir) -> tuple[Path, int]:
    zarr_path = _create_zarr(tmpdir, zarr_format=3)
    return str(zarr_path.absolute()), 3


@fixture(
    params=[
        "local_zarr_path_v2",
        # "local_zarr_path_v3",
        "local_zarr_str_v2",
        # "local_zarr_str_v3",
        # "local_zarr_store_v2",
    ]
)
def store_fixture(request):
    return request.getfixturevalue(request.param)



"""Utility functions for creating and modifying metadata."""

from typing import Any

from ngio.ngff_meta.fractal_image_meta import (
    Axis,
    Channel,
    ChannelNames,
    Dataset,
    ImageMeta,
    LabelMeta,
    Omero,
    PixelSize,
    SpaceNames,
    SpaceUnits,
    TimeNames,
    TimeUnits,
)


def _create_multiscale_meta(
    on_disk_axis: list[str] = ("t", "c", "z", "y", "x"),
    pixel_sizes: PixelSize | None = None,
    xy_scaling_factor: float = 2.0,
    z_scaling_factor: float = 1.0,
    pixel_units: SpaceUnits | str = SpaceUnits.micrometer,
    time_spacing: float = 1.0,
    time_units: TimeUnits | str = TimeUnits.s,
    num_levels: int = 5,
) -> tuple[list[Dataset], Omero]:
    """Create a image metadata object from scratch."""
    allowed_axes_names = (
        SpaceNames.allowed_names()
        + TimeNames.allowed_names()
        + ChannelNames.allowed_names()
    )
    for ax in on_disk_axis:
        if ax not in allowed_axes_names:
            raise ValueError(
                f"Invalid axis name: {ax}, allowed names: {allowed_axes_names}"
            )

    if pixel_sizes is None:
        pixel_sizes = PixelSize(z=1.0, y=1.0, x=1.0, unit=pixel_units)

    pixel_sizes_dict = pixel_sizes.as_dict()
    pixel_sizes_dict["t"] = time_spacing

    scaling_factor_dict = {
        "z": z_scaling_factor,
        "y": xy_scaling_factor,
        "x": xy_scaling_factor,
    }

    axes = Axis.batch_create(on_disk_axis, time_unit=time_units, space_unit=pixel_units)
    datasets = []
    for level in range(num_levels):
        scale = [
            pixel_sizes_dict.get(ax, 1.0) * scaling_factor_dict.get(ax, 1.0) ** level
            for ax in on_disk_axis
        ]

        datasets.append(
            Dataset(
                path=str(level),
                on_disk_axes=axes,
                on_disk_scale=scale,
                on_disk_translation=None,
            )
        )
    return datasets


def create_image_metadata(
    on_disk_axis: list[str] = ("t", "c", "z", "y", "x"),
    pixel_sizes: PixelSize | None = None,
    xy_scaling_factor: float = 2.0,
    z_scaling_factor: float = 1.0,
    time_spacing: float = 1.0,
    time_units: TimeUnits | str = TimeUnits.s,
    num_levels: int = 5,
    name: str | None = None,
    channel_labels: list[str] | None = None,
    channel_wavelengths: list[str] | None = None,
    channel_kwargs: list[dict[str, Any]] | None = None,
    omero_kwargs: dict[str, Any] | None = None,
    version: str = "0.4",
) -> ImageMeta:
    """Create a image metadata object from scratch.

    Args:
        on_disk_axis: The names of the axes. The order will correspond to the
            on-disk order. Axes order should follow the canonical order
            (t, c, z, y, x). Note that a different order can still be used
            to store the data on disk if NGFF version used is supports it.
        pixel_sizes: The pixel sizes for the z, y, x axes.
        xy_scaling_factor: The scaling factor for the y and x axes, to be used
            for the pyramid building.
        z_scaling_factor: The scaling factor for the z axis, to be used for the
            pyramid building. Note that several tools may not support scaling
            different than 1.0 for the z axis.
        time_spacing: The time spacing (If the time axis is present).
        time_units: The units of the time spacing (If the time axis is present).
        num_levels: The number of levels in the pyramid.
        name: The name of the metadata.
        channel_labels: The names of the channels.
        channel_wavelengths: The wavelengths of the channels.
        channel_kwargs: The additional channel kwargs.
        omero_kwargs: The additional omero kwargs.
        version: The version of NGFF metadata.

    """
    datasets = _create_multiscale_meta(
        on_disk_axis=on_disk_axis,
        pixel_sizes=pixel_sizes,
        xy_scaling_factor=xy_scaling_factor,
        z_scaling_factor=z_scaling_factor,
        time_spacing=time_spacing,
        time_units=time_units,
        num_levels=num_levels,
    )

    if channel_labels is None:
        return ImageMeta(
            version=version,
            name=name,
            datasets=datasets,
            omero=None,
        )

    if channel_wavelengths is None:
        channel_wavelengths = channel_labels
    else:
        if len(channel_wavelengths) != len(channel_labels):
            raise ValueError(
                "The number of channel wavelengths must match the number of "
                "channel labels."
            )

    if channel_kwargs is None:
        channel_kwargs = [{}] * len(channel_labels)
    else:
        if len(channel_kwargs) != len(channel_labels):
            raise ValueError(
                "The number of channel kwargs must match the number of channel labels."
            )

    channels = []
    for label, wavelengths, kwargs in zip(
        channel_labels, channel_wavelengths, channel_kwargs, strict=True
    ):
        channels.append(Channel(label=label, wavelength_id=wavelengths, **kwargs))

    omero_kwargs = {} if omero_kwargs is None else omero_kwargs
    omero = Omero(channels=channels, **omero_kwargs)

    return ImageMeta(
        version=version,
        name=name,
        datasets=datasets,
        omero=omero,
    )


def create_label_metadata(
    on_disk_axis: list[str] = ("t", "c", "z", "y", "x"),
    pixel_sizes: PixelSize | None = None,
    xy_scaling_factor: float = 2.0,
    z_scaling_factor: float = 1.0,
    time_spacing: float = 1.0,
    time_units: TimeUnits | str = TimeUnits.s,
    num_levels: int = 5,
    name: str | None = None,
    version: str = "0.4",
) -> LabelMeta:
    """Create a label metadata object from scratch.

    Args:
        on_disk_axis: The names of the axes. The order will correspond to the
            on-disk order. Axes order should follow the canonical order
            (t, c, z, y, x). Note that a different order can still be used
            to store the data on disk if NGFF version used is supports it.
        pixel_sizes: The pixel sizes for the z, y, x axes.
        xy_scaling_factor: The scaling factor for the y and x axes, to be used
            for the pyramid building.
        z_scaling_factor: The scaling factor for the z axis, to be used for the
            pyramid building. Note that several tools may not support scaling
            different than 1.0 for the z axis.
        time_spacing: The time spacing (If the time axis is present).
        time_units: The units of the time spacing (If the time axis is present).
        num_levels: The number of levels in the pyramid.
        name: The name of the metadata.
        version: The version of NGFF metadata.
    """
    datasets = _create_multiscale_meta(
        on_disk_axis=on_disk_axis,
        pixel_sizes=pixel_sizes,
        xy_scaling_factor=xy_scaling_factor,
        z_scaling_factor=z_scaling_factor,
        time_spacing=time_spacing,
        time_units=time_units,
        num_levels=num_levels,
    )
    return LabelMeta(
        version=version,
        name=name,
        datasets=datasets,
    )


def remove_axis_from_metadata(
    metadata: ImageMeta,
    *,
    axis_name: str | None = None,
) -> ImageMeta:
    """Remove an axis from the metadata.

    Args:
        metadata: A ImageMeta object.
        axis_name: The name of the axis to remove.
    """
    return metadata.remove_axis(axis_name=axis_name)


def add_axis_to_metadata(
    metadata: ImageMeta | LabelMeta,
    axis_name: str,
    scale: float = 1.0,
) -> ImageMeta | LabelMeta:
    """Add an axis to the ImageMeta or LabelMeta object.

    Args:
        metadata: A ImageMeta or LabelMeta object.
        axis_name: The name of the axis to add.
        scale: The scale of the axis
    """
    return metadata.add_axis(
        axis_name=axis_name,
        scale=scale,
    )


def derive_image_metadata(
    image: ImageMeta,
    name: str,
    start_level: int = 0,
) -> ImageMeta:
    """Derive a new image metadata from an existing one."""
    pass


def derive_label_metadata(
    image: ImageMeta,
    name: str,
    start_level: int = 0,
) -> LabelMeta:
    """Derive a new label metadata from an existing one."""
    pass



import json
from importlib.metadata import version
from pathlib import Path

import zarr
from packaging.version import Version
from pytest import fixture

zarr_version = version("zarr")
ZARR_PYTHON_V = 2 if Version(zarr_version) < Version("3.0.0a") else 3


@fixture
def ome_zarr_image_v04_path(tmpdir):
    zarr_path = Path(tmpdir) / "test_ome_ngff_v04.zarr"

    if ZARR_PYTHON_V == 3:
        group = zarr.open_group(store=zarr_path, mode="w", zarr_format=2)
    else:
        group = zarr.open_group(store=zarr_path, mode="w", zarr_version=2)

    json_path = (
        Path(".") / "tests" / "data" / "meta_v04" / "base_ome_zarr_image_meta.json"
    )
    with open(json_path) as f:
        base_ome_zarr_meta = json.load(f)

    base_ome_zarr_meta = base_ome_zarr_meta
    group.attrs.update(base_ome_zarr_meta)

    # shape = (3, 10, 256, 256)
    for i, path in enumerate(["0", "1", "2", "3", "4"]):
        shape = (3, 10, 256 // (2**i), 256 // (2**i))
        if ZARR_PYTHON_V == 3:
            group.create_array(name=path, fill_value=0, shape=shape)
        else:
            group.zeros(name=path, shape=shape)

    return zarr_path



"""Utility functions for creating and manipulating images."""

import math
from typing import Any

from ngio.io import StoreLike
from ngio.ngff_meta import (
    create_image_metadata,
    create_label_metadata,
    get_ngff_image_meta_handler,
)
from ngio.ngff_meta.fractal_image_meta import (
    PixelSize,
    TimeUnits,
)

try:
    from dask.distributed import Lock
except ImportError:
    Lock = None


def _build_empty_pyramid(
    group,
    image_meta,
    shape: list[int],
    chunks: list[int] | None = None,
    dtype: str = "uint16",
    on_disk_axis: list[str] = ("t", "c", "z", "y", "x"),
    xy_scaling_factor: float = 2.0,
    z_scaling_factor: float = 1.0,
):
    # Return the an Image object
    scaling_factor = []
    for ax in on_disk_axis:
        if ax in ["x", "y"]:
            scaling_factor.append(xy_scaling_factor)
        elif ax == "z":
            scaling_factor.append(z_scaling_factor)
        else:
            scaling_factor.append(1.0)

    for dataset in image_meta.datasets:
        path = dataset.path

        # V3
        # group.create_array(
        #    name=path, fill_value=0, shape=shape, dtype=dtype, chunks=chunks,
        # )

        group.zeros(
            name=path,
            shape=shape,
            dtype=dtype,
            chunks=chunks,
            dimension_separator="/",
        )

        # Todo redo this with when a proper build of pyramid is implemented
        _shape = []
        for s, sc in zip(shape, scaling_factor, strict=True):
            if math.floor(s / sc) % 2 == 0:
                _shape.append(math.floor(s / sc))
            else:
                _shape.append(math.ceil(s / sc))
        shape = list(_shape)

        if chunks is not None:
            _chunks = []
            for c, sc in zip(chunks, scaling_factor, strict=True):
                if math.floor(c / sc) % 2 == 0:
                    _chunks.append(math.floor(c / sc))
                else:
                    _chunks.append(math.ceil(c / sc))
            chunks = list(_chunks)
    return None


def create_empty_ome_zarr_image(
    store: StoreLike,
    shape: list[int],
    chunks: list[int] | None = None,
    dtype: str = "uint16",
    on_disk_axis: list[str] = ("t", "c", "z", "y", "x"),
    pixel_sizes: PixelSize | None = None,
    xy_scaling_factor: float = 2.0,
    z_scaling_factor: float = 1.0,
    time_spacing: float = 1.0,
    time_units: TimeUnits | str = TimeUnits.s,
    num_levels: int = 5,
    name: str | None = None,
    channel_labels: list[str] | None = None,
    channel_wavelengths: list[str] | None = None,
    channel_kwargs: list[dict[str, Any]] | None = None,
    omero_kwargs: dict[str, Any] | None = None,
    overwrite: bool = True,
    version: str = "0.4",
) -> None:
    """Create an empty OME-Zarr image with the given shape and metadata."""
    if len(shape) != len(on_disk_axis):
        raise ValueError(
            "The number of dimensions in the shape must match the number of "
            "axes in the on-disk axis."
        )

    if "c" in on_disk_axis:
        num_channels = shape[on_disk_axis.index("c")]
        if channel_labels is None:
            channel_labels = [f"C{i:02d}" for i in range(num_channels)]
        else:
            if len(channel_labels) != num_channels:
                raise ValueError(
                    "The number of channel labels must match the number "
                    f"of channels in the shape. Got {len(channel_labels)} "
                    f"labels for {num_channels} channels."
                )

    image_meta = create_image_metadata(
        on_disk_axis=on_disk_axis,
        pixel_sizes=pixel_sizes,
        xy_scaling_factor=xy_scaling_factor,
        z_scaling_factor=z_scaling_factor,
        time_spacing=time_spacing,
        time_units=time_units,
        num_levels=num_levels,
        name=name,
        channel_labels=channel_labels,
        channel_wavelengths=channel_wavelengths,
        channel_kwargs=channel_kwargs,
        omero_kwargs=omero_kwargs,
        version=version,
    )

    # Open the store (if it is not empty, fail)
    mode = "w" if overwrite else "w-"
    meta_handler = get_ngff_image_meta_handler(
        store=store, version=version, meta_mode="image", mode=mode
    )
    meta_handler.write_meta(image_meta)
    group = meta_handler.group

    # Create the empty image at each level in the pyramid
    _build_empty_pyramid(
        group=group,
        image_meta=image_meta,
        shape=shape,
        chunks=chunks,
        dtype=dtype,
        on_disk_axis=on_disk_axis,
        xy_scaling_factor=xy_scaling_factor,
        z_scaling_factor=z_scaling_factor,
    )


def create_empty_ome_zarr_label(
    store: StoreLike,
    shape: list[int],
    chunks: list[int] | None = None,
    dtype: str = "uint16",
    on_disk_axis: list[str] = ("t", "z", "y", "x"),
    pixel_sizes: PixelSize | None = None,
    xy_scaling_factor: float = 2.0,
    z_scaling_factor: float = 1.0,
    time_spacing: float = 1.0,
    time_units: TimeUnits | str = TimeUnits.s,
    num_levels: int = 5,
    name: str | None = None,
    overwrite: bool = True,
    version: str = "0.4",
) -> None:
    """Create an empty OME-Zarr image with the given shape and metadata."""
    if len(shape) != len(on_disk_axis):
        raise ValueError(
            "The number of dimensions in the shape must match the number of "
            "axes in the on-disk axis."
        )

    image_meta = create_label_metadata(
        on_disk_axis=on_disk_axis,
        pixel_sizes=pixel_sizes,
        xy_scaling_factor=xy_scaling_factor,
        z_scaling_factor=z_scaling_factor,
        time_spacing=time_spacing,
        time_units=time_units,
        num_levels=num_levels,
        name=name,
        version=version,
    )

    # Open the store (if it is not empty, fail)
    mode = "w" if overwrite else "w-"
    meta_handler = get_ngff_image_meta_handler(
        store=store, version=version, meta_mode="label", mode=mode
    )
    meta_handler.write_meta(image_meta)
    group = meta_handler.group
    group.attrs["image-label"] = {"version": version, "source": {"image": "../../"}}

    # Create the empty image at each level in the pyramid
    _build_empty_pyramid(
        group=group,
        image_meta=image_meta,
        shape=shape,
        chunks=chunks,
        dtype=dtype,
        on_disk_axis=on_disk_axis,
        xy_scaling_factor=xy_scaling_factor,
        z_scaling_factor=z_scaling_factor,
    )



"""Implementation of the Masking ROI Table class.

This class follows the roi_table specification at:
https://fractal-analytics-platform.github.io/fractal-tasks-core/tables/
"""

from functools import partial
from typing import Literal

import pandas as pd
import zarr
from pydantic import BaseModel

from ngio.core.label_handler import Label
from ngio.core.roi import WorldCooROI
from ngio.tables._utils import validate_columns
from ngio.tables.v1._generic_table import REQUIRED_COLUMNS, BaseTable, write_table_ad


class MaskingROITableV1Meta(BaseModel):
    """Metadata for the ROI table."""

    fractal_table_version: Literal["1"] = "1"
    type: Literal["masking_roi_table"] = "masking_roi_table"
    region: dict[Literal["path"], str]
    instance_key: str = "label"


def create_empty_roi_table() -> pd.DataFrame:
    """Create an empty ROI table."""
    table = pd.DataFrame(
        index=pd.Index([], name="label", dtype="str"), columns=REQUIRED_COLUMNS
    )

    return table


class MaskingROITableV1:
    """Class to handle fractal Masking ROI tables.

    To know more about the Masking ROI table format, please refer to the
    specification at:
    https://fractal-analytics-platform.github.io/fractal-tasks-core/tables/
    """

    def __init__(
        self,
        group: zarr.Group,
        validate_metadata: bool = True,
        validate_table: bool = True,
        index_key: str | None = None,
    ):
        """Initialize the class from an existing group.

        Args:
            group (zarr.Group): The group containing the
                ROI table.
            validate_metadata (bool): If True, the metadata is validated.
            validate_table (bool): If True, the table is validated.
            index_key (str): The column name to use as the index of the DataFrame.
        """
        if validate_metadata:
            self._meta = MaskingROITableV1Meta(**group.attrs)
        else:
            self._meta = MaskingROITableV1Meta.model_construct(**group.attrs)

        # Validate the table is not implemented for the Masking ROI table
        validators = [partial(validate_columns, required_columns=REQUIRED_COLUMNS)]
        validators = validators if validate_table else None

        index_key = index_key if index_key is not None else self._meta.instance_key
        self._table_handler = BaseTable(
            group=group,
            index_key=index_key,
            index_type="int",
            validators=validators,
        )

    @classmethod
    def _new(
        cls,
        parent_group: zarr.Group,
        name: str,
        label_image: str,
        instance_key: str = "label",
        overwrite: bool = False,
    ):
        """Create a new Masking ROI table.

        Note this method is not meant to be called directly.
        But should be called from the TablesHandler class.

        Args:
            parent_group (zarr.Group): The parent group where the ROI table
                will be created.
            name (str): The name of the ROI table.
            label_image (str): Name of the label image.
            instance_key (str): The column name to use as the index of the DataFrame.
                Default is 'label'.
            overwrite (bool): Whether to overwrite the table if it already exists.
        """
        group = parent_group.create_group(name, overwrite=overwrite)

        table = pd.DataFrame(
            index=pd.Index([], name=instance_key, dtype="int"), columns=REQUIRED_COLUMNS
        )

        meta = MaskingROITableV1Meta(
            region={"path": label_image}, instance_key=instance_key
        )
        write_table_ad(
            group=group,
            table=table,
            index_key=instance_key,
            index_type="int",
            meta=meta,
        )
        return cls(group=group)

    @property
    def meta(self) -> MaskingROITableV1Meta:
        """Return the metadata of the ROI table."""
        return self._meta

    @property
    def table_handler(self) -> BaseTable:
        """Return the table handler."""
        return self._table_handler

    @property
    def table(self) -> pd.DataFrame:
        """Return the ROI table as a DataFrame."""
        return self._table_handler.table

    @table.setter
    def table(self, table: pd.DataFrame):
        """Set the feature table."""
        raise NotImplementedError(
            "Setting the table is not implemented. Please use the 'set_table' method."
        )

    def set_table(self, table: pd.DataFrame):
        """Set the feature table."""
        self._table_handler.set_table(table)

    @property
    def list_labels(self) -> list[str]:
        """Return a list of all field indexes in the table."""
        return self.table.index.tolist()

    def from_label(self, label: Label, overwrite: bool = False) -> None:
        """Create a new ROI table from a label.

        Args:
            label (Label): The label to create the ROI table from.
            overwrite (bool): Whether to overwrite the elements in the table.
        """
        if not overwrite and self.table.empty:
            raise ValueError(
                "The table is not empty. Set overwrite to True to proceed."
            )
        raise NotImplementedError("Method not implemented yet.")

    def get_roi(self, label: int) -> WorldCooROI:
        """Get an ROI from the table."""
        if label not in self.list_labels:
            raise ValueError(f"Label {label} not found in the table.")

        table_df = self.table.loc[label]

        region_path = self.meta.region["path"]
        label_name = region_path.split("/")[-1]

        roi = WorldCooROI(
            x=table_df.loc["x_micrometer"],
            y=table_df.loc["y_micrometer"],
            z=table_df.loc["z_micrometer"],
            x_length=table_df.loc["len_x_micrometer"],
            y_length=table_df.loc["len_y_micrometer"],
            z_length=table_df.loc["len_z_micrometer"],
            unit="micrometer",
            infos={
                "label": label,
                "label_image": region_path,
                "label_name": label_name,
            },
        )
        return roi

    @property
    def rois(self) -> list[WorldCooROI]:
        """List all ROIs in the table."""
        return [self.get_roi(label) for label in self.list_labels]

    def write(self) -> None:
        """Write the crrent state of the table to the Zarr file."""
        self._table_handler.write(self.meta)



"""A module to handle OME-NGFF images stored in Zarr format."""

from ngio.core.image_like_handler import ImageLike
from ngio.io import StoreOrGroup
from ngio.ngff_meta.fractal_image_meta import ImageMeta, PixelSize


class Image(ImageLike):
    """A class to handle OME-NGFF images stored in Zarr format.

    This class provides methods to load image data and metadata from
    an OME-Zarr file.
    """

    def __init__(
        self,
        store: StoreOrGroup,
        *,
        path: str | None = None,
        idx: int | None = None,
        pixel_size: PixelSize | None = None,
        highest_resolution: bool = False,
        strict: bool = True,
        cache: bool = True,
        label_group=None,
    ) -> None:
        """Initialize the the Image Object.

        Note: Only one of `path`, `idx`, 'pixel_size' or 'highest_resolution'
        should be provided.

        store (StoreOrGroup): The Zarr store or group containing the image data.
        path (str | None): The path to the level.
        idx (int | None): The index of the level.
        pixel_size (PixelSize | None): The pixel size of the level.
        highest_resolution (bool): Whether to get the highest resolution level.
        strict (bool): Whether to raise an error where a pixel size is not found
            to match the requested "pixel_size".
        cache (bool): Whether to cache the metadata.
        label_group: The group containing the labels.
        """
        super().__init__(
            store=store,
            path=path,
            idx=idx,
            pixel_size=pixel_size,
            highest_resolution=highest_resolution,
            strict=strict,
            meta_mode="image",
            cache=cache,
            _label_group=label_group,
        )

    @property
    def metadata(self) -> ImageMeta:
        """Return the metadata of the image."""
        return super().metadata

    @property
    def channel_labels(self) -> list[str]:
        """Return the names of the channels in the image."""
        return self.metadata.channel_labels

    def get_channel_idx(
        self,
        label: str | None = None,
        wavelength_id: str | None = None,
    ) -> int:
        """Return the index of the channel."""
        return self.metadata.get_channel_idx(label=label, wavelength_id=wavelength_id)



"""Pydantic Models related to OME-NGFF 0.4 specs."""

from ngio.ngff_meta.v04.specs import NgffImageMeta04
from ngio.ngff_meta.v04.zarr_utils import NgffImageMetaZarrHandlerV04

__all__ = ["NgffImageMetaZarrHandlerV04", "NgffImageMeta04"]



class TestNgffImage:
    def test_ngff_image(self, ome_zarr_image_v04_path):
        from ngio.core.ngff_image import NgffImage

        ngff_image = NgffImage(ome_zarr_image_v04_path)
        image_handler = ngff_image.get_image(path="0")

        assert ngff_image.num_levels == 5
        assert ngff_image.levels_paths == ["0", "1", "2", "3", "4"]
        assert image_handler.channel_labels == ["DAPI", "nanog", "Lamin B1"]
        assert image_handler.get_channel_idx(label="DAPI") == 0
        assert image_handler.get_channel_idx(wavelength_id="A01_C01") == 0

        new_path = ome_zarr_image_v04_path.parent / "new_ngff_image.zarr"
        new_ngff_image = ngff_image.derive_new_image(
            new_path, "new_image", overwrite=True
        )
        new_image_handler = new_ngff_image.get_image(path="0")

        assert new_ngff_image.levels_paths == ngff_image.levels_paths
        assert new_image_handler.channel_labels == image_handler.channel_labels
        assert new_image_handler.shape == image_handler.shape
        assert new_image_handler.pixel_size.zyx == image_handler.pixel_size.zyx
        assert (
            new_image_handler.on_disk_array.shape == image_handler.on_disk_array.shape
        )
        assert (
            new_image_handler.on_disk_array.chunks == image_handler.on_disk_array.chunks
        )



"""Image metadata models.

This module contains the models for the image metadata.
These metadata models are not adhering to the OME standard.
But they can be built from the OME standard metadata, and the
can be converted to the OME standard.
"""

from enum import Enum

import numpy as np
from pydantic import BaseModel, Field
from typing_extensions import Self

from ngio.pydantic_utils import BaseWithExtraFields


class Channel(BaseWithExtraFields):
    """Information about a channel in the image.

    Attributes:
        label(str): The label of the channel.
        wavelength_id(str): The wavelength ID of the channel.
        extra_fields(dict): To reduce the api surface, extra fields are stored in the
            the channel attributes will be stored in the extra_fields attribute.
    """

    label: str
    wavelength_id: str | None = None


class Omero(BaseWithExtraFields):
    """Information about the OMERO metadata.

    Attributes:
        channels(list[Channel]): The list of channels in the image.
        extra_fields(dict): To reduce the api surface, extra fields are stored in the
            the omero attributes will be stored in the extra_fields attribute.
    """

    channels: list[Channel] = Field(default_factory=list)


class AxisType(str, Enum):
    """Allowed axis types."""

    channel = "channel"
    time = "time"
    space = "space"


class SpaceUnits(str, Enum):
    """Allowed space units."""

    nanometer = "nanometer"
    nm = "nm"
    micrometer = "micrometer"
    um = "um"
    millimeter = "millimeter"
    mm = "mm"
    centimeter = "centimeter"
    cm = "cm"

    @classmethod
    def allowed_names(self) -> list[str]:
        """Get the allowed space axis names."""
        return list(SpaceUnits.__members__.keys())


class SpaceNames(str, Enum):
    """Allowed space axis names."""

    z = "z"
    y = "y"
    x = "x"

    @classmethod
    def allowed_names(self) -> list[str]:
        """Get the allowed space axis names."""
        return list(SpaceNames.__members__.keys())


class ChannelNames(str, Enum):
    """Allowed channel axis names."""

    c = "c"

    @classmethod
    def allowed_names(self) -> list[str]:
        """Get the allowed channel axis names."""
        return list(ChannelNames.__members__.keys())


class PixelSize(BaseModel):
    """PixelSize class to store the pixel size in 3D space."""

    x: float = Field(..., ge=0)
    y: float = Field(..., ge=0)
    z: float = Field(1.0, ge=0)
    unit: SpaceUnits = SpaceUnits.micrometer
    virtual: bool = False

    def __repr__(self):
        """Return the string representation of the object."""
        return f"PixelSize(x={self.x}, y={self.y}, z={self.z}, unit={self.unit.value})"

    @classmethod
    def from_list(cls, sizes: list[float], unit: SpaceUnits):
        """Build a PixelSize object from a list of sizes.

        Note: The order of the sizes must be z, y, x.

        Args:
            sizes(list[float]): The list of sizes.
            unit(SpaceUnits): The unit of the sizes.
        """
        if len(sizes) == 2:
            return cls(y=sizes[0], x=sizes[1], unit=unit)
        elif len(sizes) == 3:
            return cls(z=sizes[0], y=sizes[1], x=sizes[2], unit=unit)
        else:
            raise ValueError("Invalid pixel size list. Must have 2 or 3 elements.")

    def as_dict(self) -> dict:
        """Return the pixel size as a dictionary."""
        return {"z": self.z, "y": self.y, "x": self.x}

    @property
    def zyx(self) -> tuple:
        """Return the voxel size in z, y, x order."""
        return self.z, self.y, self.x

    @property
    def yx(self) -> tuple:
        """Return the xy plane pixel size in y, x order."""
        return self.y, self.x

    def voxel_volume(self) -> float:
        """Return the volume of a voxel."""
        return self.y * self.x * (self.z or 1)

    def xy_plane_area(self) -> float:
        """Return the area of the xy plane."""
        return self.y * self.x

    def distance(self, other: "PixelSize") -> float:
        """Return the distance between two pixel sizes."""
        return np.linalg.norm(np.array(self.zyx) - np.array(other.zyx))


class TimeUnits(str, Enum):
    """Allowed time units."""

    seconds = "seconds"
    s = "s"

    @classmethod
    def allowed_names(self) -> list[str]:
        """Get the allowed time axis names."""
        return list(TimeUnits.__members__.keys())


class TimeNames(str, Enum):
    """Allowed time axis names."""

    t = "t"

    @classmethod
    def allowed_names(self) -> list[str]:
        """Get the allowed time axis names."""
        return list(TimeNames.__members__.keys())


class Axis:
    """Axis infos model."""

    def __init__(
        self,
        name: str | TimeNames | SpaceNames,
        unit: SpaceUnits | TimeUnits | None = None,
    ) -> None:
        """Initialize the Axis object.

        Args:
            name(str): The name of the axis.
            unit(SpaceUnits | TimeUnits | None): The unit of the axis.
        """
        if name is None:
            raise ValueError("Axis name cannot be None.")

        if isinstance(name, Enum):
            name = name.value

        self._name = name

        if name in TimeNames.allowed_names():
            self._type = AxisType.time

            if unit is None:
                self._unit = TimeUnits.s
            elif unit not in TimeUnits.allowed_names():
                raise ValueError(f"Invalid time unit {unit}.")
            else:
                self._unit = unit

        elif name in SpaceNames.allowed_names():
            self._type = AxisType.space

            if unit is None:
                self._unit = SpaceUnits.um
            elif unit not in SpaceUnits.allowed_names():
                raise ValueError(f"Invalid space unit {unit}.")
            else:
                self._unit = unit

        elif name in ChannelNames.allowed_names():
            self._type = AxisType.channel
            if unit is not None:
                raise ValueError("Channel axis cannot have a unit.")
            self._unit = None
        else:
            raise ValueError(f"Invalid axis name {name}.")

    @classmethod
    def lazy_create(
        cls,
        name: str | TimeNames | SpaceNames,
        time_unit: TimeUnits | None = None,
        space_unit: SpaceUnits | None = None,
    ) -> "Axis":
        """Create an Axis object with the default unit."""
        if name in TimeNames.allowed_names():
            return cls(name=name, unit=time_unit)
        elif name in SpaceNames.allowed_names():
            return cls(name=name, unit=space_unit)
        else:
            return cls(name=name, unit=None)

    @classmethod
    def batch_create(
        cls,
        axes_names: list[str | SpaceNames | TimeNames],
        time_unit: TimeUnits | None = None,
        space_unit: SpaceUnits | None = None,
    ) -> list["Axis"]:
        """Create a list of Axis objects from a list of dictionaries."""
        axes = []
        for name in axes_names:
            axes.append(
                cls.lazy_create(name=name, time_unit=time_unit, space_unit=space_unit)
            )
        return axes

    @property
    def name(self) -> str:
        """Get the name of the axis."""
        return self._name

    @property
    def unit(self) -> SpaceUnits | TimeUnits | None:
        """Get the unit of the axis."""
        return self._unit

    @unit.setter
    def unit(self, unit: SpaceUnits | TimeUnits | None):
        """Set the unit of the axis."""
        self._unit = unit

    @property
    def type(self) -> AxisType:
        """Get the type of the axis."""
        return self._type

    def model_dump(self) -> dict:
        """Return the axis information in a dictionary."""
        return {"name": self.name, "unit": self.unit, "type": self.type}


class Dataset:
    """Model for a dataset in the multiscale.

    To initialize the Dataset object, the path, the axes, scale, and translation list
    can be provided with on_disk order.

    The Dataset object will reorder the scale and translation lists according to the
    following canonical order of the axes:
        * Time axis (if present)
        * Channel axis (if present)
        * Z axis (if present)
        * Y axis (Mandatory)
        * X axis (Mandatory)
    """

    def __init__(
        self,
        *,
        path: str,
        on_disk_axes: list[Axis],
        on_disk_scale: list[float],
        on_disk_translation: list[float] | None = None,
        canonical_order: list[str] | None = None,
    ):
        """Initialize the Dataset object.

        Args:
            path(str): The path of the dataset.
            on_disk_axes(list[Axis]): The list of axes in the multiscale.
            on_disk_scale(list[float]): The list of scale transformation.
                The scale transformation must have the same length as the axes.
            on_disk_translation(list[float] | None): The list of translation.
            canonical_order(list[str] | None): The canonical order of the axes.
                If None, the default order is ["t", "c", "z", "y", "x"].
        """
        self._path = path

        # Canonical order validation
        if canonical_order is None:
            self._canonical_order = ["t", "c", "z", "y", "x"]
        else:
            self._canonical_order = canonical_order

        for ax in on_disk_axes:
            if ax.name not in self._canonical_order:
                raise ValueError(f"Axis {ax.name} not found in the canonical order.")

        if len(set(self._canonical_order)) != len(self._canonical_order):
            raise ValueError("Canonical order must have unique elements.")

        if len(set(on_disk_axes)) != len(on_disk_axes):
            raise ValueError("on_disk axes must have unique elements.")

        self._on_disk_axes = on_disk_axes

        # Scale transformation validation
        if len(on_disk_scale) != len(on_disk_axes):
            raise ValueError(
                "Inconsistent scale transformation. "
                "The scale transformation must have the same length."
            )
        self._scale = on_disk_scale

        # Translation transformation validation
        if on_disk_translation is not None and len(on_disk_translation) != len(
            on_disk_axes
        ):
            raise ValueError(
                "Inconsistent translation transformation. "
                "The translation transformation must have the same length."
            )

        self._translation = on_disk_translation

        # Compute the index mapping between the canonical order and the actual order
        _map = {ax.name: i for i, ax in enumerate(on_disk_axes)}
        self._index_mapping = {
            name: _map.get(name, None) for name in self._canonical_order
        }
        self._ordered_axes = [
            on_disk_axes[i] for i in self._index_mapping.values() if i is not None
        ]

    @property
    def path(self) -> str:
        """Get the path of the dataset."""
        return self._path

    @property
    def index_mapping(self) -> dict[str, int]:
        """Get the mapping between the canonical order and the actual order."""
        return self._index_mapping

    @property
    def axes(self) -> list[Axis]:
        """Get the axes in the canonical order."""
        return self._ordered_axes

    @property
    def on_disk_axes_names(self) -> list[Axis]:
        """Get the axes in the on-disk order."""
        return [ax.name for ax in self._on_disk_axes]

    @property
    def axes_order(self) -> list[str]:
        """Get the mapping between the canonical order and the on-disk order.

        Example:
            on_disk_order = ["z", "c", "y", "x"]
            canonical_order = ["c", "z", "y", "x"]
            axes_order = [1, 0, 2, 3]
        """
        on_disk_axes = self.on_disk_axes_names
        canonical_axes = self.axes_names
        return [on_disk_axes.index(ax) for ax in canonical_axes]

    @property
    def reverse_axes_order(self) -> list[str]:
        """Get the mapping between the on-disk order and the canonical order.

        It is the inverse of the axes_order.
        """
        return np.argsort(self.axes_order)

    @property
    def scale(self) -> list[float]:
        """Get the scale transformation of the dataset in the canonical order."""
        return [self._scale[i] for i in self._index_mapping.values() if i is not None]

    @property
    def time_spacing(self) -> float:
        """Get the time spacing of the dataset."""
        if "t" not in self.axes_names:
            return 1.0

        scale_t = self.scale[self.index_mapping.get("t")]
        return scale_t

    @property
    def on_disk_scale(self) -> list[float]:
        """Get the scale transformation of the dataset in the on-disk order."""
        return self._scale

    @property
    def translation(self) -> list[float] | None:
        """Get the translation transformation of the dataset in the canonical order."""
        if self._translation is None:
            return None
        return [self._translation[i] for i in self._index_mapping.values()]

    @property
    def axes_names(self) -> list[str]:
        """Get the axes names in the canonical order."""
        return [ax.name for ax in self.axes]

    @property
    def space_axes_names(self) -> list[str]:
        """Get the spatial axes names in the canonical order."""
        return [ax.name for ax in self.axes if ax.type == AxisType.space]

    @property
    def space_axes_unit(self) -> SpaceUnits:
        """Get the unit of the spatial axes."""
        types = [ax.unit for ax in self.axes if ax.type == AxisType.space]
        if len(set(types)) > 1:
            raise ValueError("Inconsistent spatial axes units.")
        return types[0]

    @property
    def pixel_size(self) -> PixelSize:
        """Get the pixel size of the dataset."""
        pixel_sizes = {}

        for ax, scale in zip(self.axes, self.scale, strict=True):
            if ax.type == AxisType.space:
                pixel_sizes[ax.name] = scale

        return PixelSize(**pixel_sizes, unit=self.space_axes_unit)

    @property
    def time_axis_unit(self) -> TimeUnits | None:
        """Get the unit of the time axis."""
        types = [ax.unit for ax in self.axes if ax.type == AxisType.time]
        if len(set(types)) > 1:
            raise ValueError("Inconsistent time axis units.")
        return types[0] if types else None

    def remove_axis(self, axis_name: str) -> "Dataset":
        """Remove an axis from the dataset.

        Args:
            axis_name(str): The name of the axis to remove.
        """
        if axis_name not in self.axes_names:
            raise ValueError(f"Axis {axis_name} not found in the dataset.")

        if axis_name in ["x", "y"]:
            raise ValueError("Cannot remove mandatory axes x and y.")

        axes_idx = self.index_mapping[axis_name]

        new_on_disk_axes = self._on_disk_axes.copy()
        new_on_disk_axes.pop(axes_idx)

        new_scale = self._scale.copy()
        new_scale.pop(axes_idx)

        if self._translation is not None:
            new_translation = self._translation.copy()
            new_translation.pop(axes_idx)
        else:
            new_translation = None

        return Dataset(
            path=self.path,
            on_disk_axes=new_on_disk_axes,
            on_disk_scale=new_scale,
            on_disk_translation=new_translation,
            canonical_order=self._canonical_order,
        )

    def add_axis(
        self, axis_name: str, scale: float = 1.0, translation: float | None = None
    ) -> "Dataset":
        """Add an axis to the dataset.

        Args:
            axis_name(str): The name of the axis to add.
            scale(float): The scale of the axis.
            translation(float | None): The translation of the axis.
        """
        if axis_name in self.axes_names:
            raise ValueError(f"Axis {axis_name} already exists in the dataset.")

        axis = Axis.lazy_create(
            name=axis_name,
            space_unit=self.space_axes_unit,
            time_unit=self.time_axis_unit,
        )

        new_on_disk_axes = self._on_disk_axes.copy()
        new_on_disk_axes.append(axis)

        new_scale = self._scale.copy()
        new_scale.append(scale)

        if self._translation is not None:
            new_translation = self._translation.copy()
            new_translation.append(translation)
        else:
            new_translation = None

        return Dataset(
            path=self.path,
            on_disk_axes=new_on_disk_axes,
            on_disk_scale=new_scale,
            on_disk_translation=new_translation,
            canonical_order=self._canonical_order,
        )

    def to_canonical_order(self) -> "Dataset":
        """Return a new Dataset where the axes are in the canonical order."""
        new_axes = self._ordered_axes
        new_scale = self.scale
        new_translation = self.translation
        return Dataset(
            path=self.path,
            on_disk_axes=new_axes,
            on_disk_scale=new_scale,
            on_disk_translation=new_translation,
        )

    def on_disk_model_dump(self) -> dict:
        """Return the dataset information in the on_disk order."""
        return {
            "path": self.path,
            "axes": [ax.model_dump(exclude_none=True) for ax in self._on_disk_axes],
            "scale": self._scale,
            "translation": self._translation,
        }

    def ordered_model_dump(self) -> dict:
        """Return the dataset information in the canonical order."""
        return {
            "path": self.path,
            "axes": [ax.model_dump(exclude_none=True) for ax in self.axes],
            "scale": self.scale,
            "translation": self.translation,
        }


class BaseMeta:
    """Base class for ImageMeta and LabelMeta."""

    def __init__(self, version: str, name: str, datasets: list[Dataset]) -> None:
        """Initialize the ImageMeta object."""
        self._version = version
        self._name = name

        if len(datasets) == 0:
            raise ValueError("At least one dataset must be provided.")

        self._datasets = datasets

    @property
    def version(self) -> str:
        """Version of the OME-NFF metadata used to build the object."""
        return self._version

    @property
    def name(self) -> str:
        """Name of the image."""
        return self._name

    @property
    def datasets(self) -> list[Dataset]:
        """List of datasets in the multiscale."""
        return self._datasets

    @property
    def num_levels(self) -> int:
        """Number of levels in the multiscale."""
        return len(self.datasets)

    @property
    def levels_paths(self) -> list[str]:
        """List of paths of the datasets."""
        return [dataset.path for dataset in self.datasets]

    @property
    def index_mapping(self) -> dict[str, int]:
        """Get the mapping between the canonical order and the actual order."""
        return self.datasets[0].index_mapping

    @property
    def axes(self) -> list[Axis]:
        """List of axes in the canonical order."""
        return self.datasets[0].axes

    @property
    def axes_names(self) -> list[str]:
        """List of axes names in the canonical order."""
        return self.datasets[0].axes_names

    @property
    def space_axes_names(self) -> list[str]:
        """List of spatial axes names in the canonical order."""
        return self.datasets[0].space_axes_names

    @property
    def space_axes_unit(self) -> SpaceUnits:
        """Get the unit of the spatial axes."""
        return self.datasets[0].space_axes_unit

    @property
    def time_axis_unit(self) -> TimeUnits | None:
        """Get the unit of the time axis."""
        return self.datasets[0].time_axis_unit

    def _get_dataset_by_path(self, path: str) -> Dataset:
        """Get a dataset by its path."""
        for dataset in self.datasets:
            if dataset.path == path:
                return dataset
        raise ValueError(f"Dataset with path {path} not found.")

    def _get_dataset_by_index(self, idx: int) -> Dataset:
        """Get a dataset by its index."""
        if idx < 0 or idx >= len(self.datasets):
            raise ValueError(f"Index {idx} out of range.")
        return self.datasets[idx]

    def _get_dataset_by_pixel_size(
        self, pixel_size: PixelSize, strict: bool = False, tol: float = 1e-6
    ) -> Dataset:
        """Get a dataset with the closest pixel size.

        Args:
            pixel_size(PixelSize): The pixel size to search for.
            strict(bool): If True, the pixel size must smaller than tol.
            tol(float): Any pixel size with a distance less than tol will be considered.
        """
        min_dist = np.inf

        for dataset in self.datasets:
            dist = dataset.pixel_size.distance(pixel_size)
            if dist < min_dist:
                min_dist = dist
                closest_dataset = dataset

        if strict and min_dist > tol:
            raise ValueError("No dataset with a pixel size close enough.")

        return closest_dataset

    def get_dataset(
        self,
        *,
        path: str | None = None,
        idx: int | None = None,
        pixel_size: PixelSize | None = None,
        highest_resolution: bool = False,
        strict: bool = False,
    ) -> Dataset:
        """Get a dataset by its path, index or pixel size.

        Args:
            path(str): The path of the dataset.
            idx(int): The index of the dataset.
            pixel_size(PixelSize): The pixel size to search for.
            highest_resolution(bool): If True, the dataset with the highest resolution
            strict(bool): If True, the pixel size must be exactly the same.
                If pixel_size is None, strict is ignored.
        """
        # Only one of the arguments must be provided
        if (
            sum(
                [
                    path is not None,
                    idx is not None,
                    pixel_size is not None,
                    highest_resolution,
                ]
            )
            != 1
        ):
            raise ValueError("get_dataset must receive only one argument.")

        if path is not None:
            return self._get_dataset_by_path(path)
        elif idx is not None:
            return self._get_dataset_by_index(idx)
        elif pixel_size is not None:
            return self._get_dataset_by_pixel_size(pixel_size, strict=strict)
        elif highest_resolution:
            return self.get_highest_resolution_dataset()
        else:
            raise ValueError("get_dataset has no valid arguments.")

    def get_highest_resolution_dataset(self) -> Dataset:
        """Get the dataset with the highest resolution."""
        return self._get_dataset_by_pixel_size(
            pixel_size=PixelSize(x=0.0, y=0.0, z=0.0, unit=SpaceUnits.um), strict=False
        )

    def scale(self, path: str | None = None, idx: int | None = None) -> list[float]:
        """Get the scale transformation of a dataset.

        Args:
            path(str): The path of the dataset.
            idx(int): The index of the dataset.
        """
        return self.get_dataset(path=path, idx=idx).scale

    def _scaling_factors(self) -> list[float]:
        scaling_factors = []
        for d1, d2 in zip(self.datasets[1:], self.datasets[:-1], strict=True):
            scaling_factors.append(
                [d1 / d2 for d1, d2 in zip(d1.scale, d2.scale, strict=True)]
            )

        for sf in scaling_factors:
            assert (
                sf == scaling_factors[0]
            ), "Inconsistent scaling factors not well supported."
        return scaling_factors[0]

    @property
    def xy_scaling_factor(self) -> float:
        """Get the xy scaling factor of the dataset."""
        scaling_factors = self._scaling_factors()
        x_scaling_f = scaling_factors[self.index_mapping.get("x")]
        y_scaling_f = scaling_factors[self.index_mapping.get("y")]

        if not np.allclose(x_scaling_f, y_scaling_f):
            raise ValueError("Inconsistent xy scaling factor.")
        return x_scaling_f

    @property
    def z_scaling_factor(self) -> float:
        """Get the z scaling factor of the dataset."""
        scaling_factors = self._scaling_factors()
        if "z" not in self.axes_names:
            return 1.0
        z_scaling_f = scaling_factors[self.index_mapping.get("z")]
        return z_scaling_f

    def translation(
        self, path: str | None = None, idx: int | None = None
    ) -> list[float] | None:
        """Get the translation transformation of a dataset.

        Args:
            path(str): The path of the dataset.
            idx(int): The index of the dataset.
        """
        return self.get_dataset(path=path, idx=idx).translation

    def pixel_size(self, path: str | None = None, idx: int | None = None) -> PixelSize:
        """Get the pixel size of a dataset.

        Args:
            path(str): The path of the dataset.
            idx(int): The index of the dataset.
        """
        return self.get_dataset(path=path, idx=idx).pixel_size

    def remove_axis(self, axis_name: str) -> Self:
        """Remove an axis from the metadata.

        Args:
            axis_name(str): The name of the axis to remove.
        """
        new_datasets = [dataset.remove_axis(axis_name) for dataset in self.datasets]
        return self.__class__(
            version=self.version, name=self.name, datasets=new_datasets
        )

    def add_axis(
        self, axis_name: str, scale: float = 1.0, translation: float | None = None
    ) -> Self:
        """Add an axis to the metadata.

        Args:
            axis_name(str): The name of the axis to add.
            scale(float): The scale of the axis.
            translation(float | None): The translation of the axis.
        """
        new_datasets = [
            dataset.add_axis(axis_name, scale, translation) for dataset in self.datasets
        ]
        return self.__class__(
            version=self.version, name=self.name, datasets=new_datasets
        )


class LabelMeta(BaseMeta):
    """Label metadata model."""

    def __init__(self, version: str, name: str, datasets: list[Dataset]) -> None:
        """Initialize the ImageMeta object."""
        super().__init__(version, name, datasets)

        # Make sure that there are no channel axes
        for ax in self.datasets[0].axes:
            if ax.type == AxisType.channel:
                raise ValueError("Channel axes are not allowed in ImageMeta.")

    def add_axis(
        self, axis_name: str, scale: float = 1, translation: float | None = None
    ) -> BaseMeta:
        """Add an axis to the metadata."""
        # Check if the axis is a channel
        axis = Axis.lazy_create(
            name=axis_name,
            space_unit=self.space_axes_unit,
            time_unit=self.time_axis_unit,
        )
        if axis.type == AxisType.channel:
            raise ValueError("Channel axes are not allowed in LabelMeta.")

        return super().add_axis(
            axis_name=axis_name, scale=scale, translation=translation
        )


class ImageMeta(BaseMeta):
    """Image metadata model."""

    def __init__(
        self,
        version: str,
        name: str,
        datasets: list[Dataset],
        omero: Omero | None = None,
    ) -> None:
        """Initialize the ImageMeta object."""
        super().__init__(version=version, name=name, datasets=datasets)
        self._omero = omero

    @property
    def omero(self) -> Omero | None:
        """Get the OMERO metadata."""
        return self._omero

    @property
    def channels(self) -> list[Channel]:
        """Get the channels in the image."""
        if self._omero is None:
            return []
        return self.omero.channels

    @property
    def num_channels(self) -> int:
        """Get the number of channels in the image."""
        return len(self.channels)

    @property
    def channel_labels(self) -> list[str]:
        """Get the labels of the channels in the image."""
        return [channel.label for channel in self.channels]

    @property
    def channel_wavelength_ids(self) -> list[str]:
        """Get the wavelength IDs of the channels in the image."""
        return [channel.wavelength_id for channel in self.channels]

    def _get_channel_idx_by_label(self, label: str) -> int | None:
        """Get the index of a channel by its label."""
        if self._omero is None:
            return None

        if label not in self.channel_labels:
            raise ValueError(f"Channel with label {label} not found.")

        return self.channel_labels.index(label)

    def _get_channel_idx_by_wavelength_id(self, wavelength_id: str) -> int | None:
        """Get the index of a channel by its wavelength ID."""
        if self._omero is None:
            return None

        if wavelength_id not in self.channel_wavelength_ids:
            raise ValueError(f"Channel with wavelength ID {wavelength_id} not found.")

        return self.channel_wavelength_ids.index(wavelength_id)

    def get_channel_idx(
        self, label: str | None = None, wavelength_id: str | None = None
    ) -> int:
        """Get the index of a channel by its label or wavelength ID."""
        # Only one of the arguments must be provided
        if sum([label is not None, wavelength_id is not None]) != 1:
            raise ValueError("get_channel_idx must receive only one argument.")

        if label is not None:
            return self._get_channel_idx_by_label(label)
        elif wavelength_id is not None:
            return self._get_channel_idx_by_wavelength_id(wavelength_id)
        else:
            raise ValueError(
                "get_channel_idx must receive either label or wavelength_id."
            )

    def remove_axis(self, axis_name: str) -> "ImageMeta":
        """Remove an axis from the metadata.

        Args:
            axis_name(str): The name of the axis to remove.
        """
        new_image = super().remove_axis(axis_name=axis_name)

        # If the removed axis is a channel, remove the channel from the omero metadata
        if axis_name in ChannelNames.allowed_names():
            new_omero = Omero(channels=[], **self.omero.extra_fields)
            return ImageMeta(
                version=new_image.version,
                name=new_image.name,
                datasets=new_image.datasets,
                omero=new_omero,
            )

        return new_image


ImageLabelMeta = ImageMeta | LabelMeta



"""A module to handle data transforms for image data."""

from ngio.pipes._slicer_transforms import NaiveSlicer, RoiSlicer
from ngio.pipes.data_pipe import DataTransformPipe

__all__ = ["DataTransformPipe", "NaiveSlicer", "RoiSlicer"]



"""Implementations of the OME-NGFF 0.4 specs using Pydantic models.

See https://ngff.openmicroscopy.org/0.4/ for detailed specs.
"""

from typing import Literal

from pydantic import BaseModel, Field, field_validator

from ngio.pydantic_utils import (
    SKIP_NGFF_VALIDATION,
    BaseWithExtraFields,
    unique_items_validator,
)


class Window04(BaseModel):
    """Model for `Channels.Window`."""

    max: int | float
    min: int | float
    start: int | float
    end: int | float


class Channel04(BaseWithExtraFields):
    """Model for `Omero.channels`."""

    active: bool = True
    # coefficient: int | None = None
    color: str
    # family: str | None = None
    # inverted: bool | None = None
    label: str | None = None
    window: Window04 | None = None


class Omero04(BaseWithExtraFields):
    """Model for `NgffImageMeta.Omero`."""

    channels: list[Channel04]
    version: Literal["0.4"] = "0.4"


class Axis04(BaseModel):
    """Model for `Multiscale.axes` elements."""

    name: str
    type: str
    unit: str | None = None


class ScaleCoordinateTransformation04(BaseModel):
    """Model for a scale transformation.

    This corresponds to scale-type elements of
    `Dataset.coordinateTransformations` or
    `Multiscale.coordinateTransformations`.
    """

    type: Literal["scale"]
    scale: list[float] = Field(..., min_length=2)


class TranslationCoordinateTransformation04(BaseModel):
    """Model for a translation transformation.

    This corresponds to translation-type elements of
    `Dataset.coordinateTransformations` or
    `Multiscale.coordinateTransformations`.
    """

    type: Literal["translation"]
    translation: list[float] = Field(..., min_length=2)


Transformation04 = (
    ScaleCoordinateTransformation04 | TranslationCoordinateTransformation04
)


class Dataset04(BaseModel):
    """Model for `Multiscale.datasets` elements."""

    path: str
    coordinateTransformations: list[Transformation04] = Field(
        ..., min_length=1, max_length=2
    )

    @field_validator("coordinateTransformations")
    @classmethod
    def _check_scale_exists(cls, v):
        # check if at least one scale transformation exists
        if SKIP_NGFF_VALIDATION:
            return v

        num_scale = sum(
            1 for item in v if isinstance(item, ScaleCoordinateTransformation04)
        )
        if num_scale != 1:
            raise ValueError("Exactly one scale transformation is required.")

        num_translation = sum(
            1 for item in v if isinstance(item, TranslationCoordinateTransformation04)
        )
        if num_translation > 1:
            raise ValueError("At most one translation transformation is allowed.")

        return v


class Multiscale04(BaseModel):
    """Model for `NgffImageMeta.multiscales` elements."""

    name: str | None = None
    datasets: list[Dataset04] = Field(..., min_length=1)
    version: Literal["0.4"] | None = "0.4"
    axes: list[Axis04] = Field(..., max_length=5, min_length=2)
    coordinateTransformations: list[Transformation04] | None = None
    _check_unique = field_validator("axes")(unique_items_validator)

    @field_validator("axes")
    @classmethod
    def _check_axes_order(cls, v):
        # check if the order of axes is correct
        if SKIP_NGFF_VALIDATION:
            return v

        axes_types = [axis.type for axis in v]

        if "time" in axes_types:
            time_position = axes_types.index("time")
            if time_position != 0:
                raise ValueError("Time axis should be the first axis.")

            axes_types = axes_types.pop(0)

        if len(axes_types) < 2:
            raise ValueError("At least two spatial axes are required.")

        reversed_axes_types = axes_types[::-1]

        channel_type_flag = False
        for ax_type in reversed_axes_types:
            if ax_type == "space":
                if channel_type_flag:
                    raise ValueError("Channel axis should precede spatial axes.")
            else:
                channel_type_flag = True
        return v


class NgffImageMeta04(BaseWithExtraFields):
    """Model for the metadata of a NGFF image."""

    multiscales: list[Multiscale04] = Field(
        ...,
        min_length=1,
    )
    omero: Omero04 | None = None
    _check_unique = field_validator("multiscales")(unique_items_validator)



from collections.abc import Callable
from typing import Literal

import anndata as ad
import numpy as np
import pandas as pd
import pandas.api.types as ptypes


class TableValidationError(Exception):
    """Error raised when a table is not formatted correctly."""

    pass


Validator = Callable[[pd.DataFrame], pd.DataFrame]


def _check_for_mixed_types(series: pd.Series) -> None:
    """Check if the column has mixed types."""
    if series.apply(type).nunique() > 1:
        raise TableValidationError(
            f"Column {series.name} has mixed types: "
            f"{series.apply(type).unique()}. "
            "Type of all elements must be the same."
        )


def _check_for_supported_types(series: pd.Series) -> Literal["str", "int", "numeric"]:
    """Check if the column has supported types."""
    if ptypes.is_string_dtype(series):
        return "str"
    if ptypes.is_integer_dtype(series):
        return "int"
    if ptypes.is_numeric_dtype(series):
        return "numeric"
    raise TableValidationError(
        f"Column {series.name} has unsupported type: {series.dtype}."
        " Supported types are string and numerics."
    )


def _check_index_key(
    table_df: pd.DataFrame, index_key: str, index_type: Literal["str", "int"]
) -> pd.DataFrame:
    """Check if the index_key correctness.

    - Check if the index_key is present in the data frame.
        (If the index_key is a column in the DataFrame, it is set as the index)
    - Check if the index_key is of the correct type.

    Args:
        table_df (pd.DataFrame): The DataFrame to validate.
        index_key (str): The column name to use as the index of the DataFrame.
        index_type (str): The type of the index column in the DataFrame.
            Either 'str' or 'int'. Default is 'int'.

    Returns:
        pd.DataFrame: The validated DataFrame.
    """
    columns = table_df.columns
    if index_key in columns:
        table_df = table_df.set_index(index_key)

    if table_df.index.name != index_key:
        raise TableValidationError(f"index_key: {index_key} not found in data frame")

    if index_type == "str":
        if ptypes.is_integer_dtype(table_df.index):
            # Convert the int index to string is generally safe
            table_df.index = table_df.index.astype(str)

        if not ptypes.is_string_dtype(table_df.index):
            raise TableValidationError(f"index_key {index_key} must be of string type")

    elif index_type == "int":
        if ptypes.is_string_dtype(table_df.index):
            # Try to convert the string index to int
            try:
                table_df.index = table_df.index.astype(int)
            except ValueError as e:
                if "invalid literal for int() with base 10" in str(e):
                    raise TableValidationError(
                        f"index_key {index_key} must be of "
                        "integer type, but found string. We "
                        "tried implicit conversion failed."
                    ) from None
                else:
                    raise e from e

        if not ptypes.is_integer_dtype(table_df.index):
            raise TableValidationError(f"index_key {index_key} must be of integer type")

    else:
        raise TableValidationError(f"index_type {index_type} not recognized")

    return table_df


def validate_table(
    table_df: pd.DataFrame,
    index_key: str,
    index_type: Literal["str", "int"],
    validators: list[Validator] | None,
) -> pd.DataFrame:
    """Validate the table DataFrame.

    - Check if the index_key is present in the data frame.
        (If the index_key is a column in the DataFrame, it is set as the index)
    - Check if the index_key is of the correct type.
    - Apply all provided optional validators.

    Args:
        table_df (pd.DataFrame): The DataFrame to validate.
        index_key (str): The column name to use as the index of the DataFrame.
        index_type (str): The type of the index column in the DataFrame.
        validators (list[Validator]): A list of functions to further validate table.

    Returns:
        pd.DataFrame: The validated DataFrame.
    """
    table_df = _check_index_key(table_df, index_key, index_type)

    if validators is None:
        return table_df

    # Apply all provided validators
    for validator in validators:
        table_df = validator(table_df)

    return table_df


def table_df_to_ad(
    table_df: pd.DataFrame,
    index_key: str,
    index_type: Literal["str", "int"] = "int",
    validators: list[Validator] | None = None,
) -> ad.AnnData:
    """Convert a table DataFrame to an AnnData object.

    Args:
        table_df (pd.DataFrame): A pandas DataFrame representing a fractal table.
        index_key (str): The column name to use as the index of the DataFrame.
        index_type (str): The type of the index column in the DataFrame.
            Either 'str' or 'int'. Default is 'int'.
        validators (list[Validator]): A list of functions to further validate the table.
    """
    # Check if the index_key is present in the data frame + optional validations
    table_df = validate_table(
        table_df=table_df,
        index_key=index_key,
        index_type=index_type,
        validators=validators,
    )

    # DO NOT SKIP
    # Convert the index to string ALWAYS to avoid casting issues in AnnData
    table_df.index = table_df.index.astype(str)

    str_columns, int_columns, num_columns = [], [], []
    for c_name in table_df.columns:
        column_df = table_df[c_name]
        _check_for_mixed_types(column_df)  # Mixed types are not allowed in the table
        c_type = _check_for_supported_types(
            column_df
        )  # Only string and numeric types are allowed

        if c_type == "str":
            str_columns.append(c_name)

        elif c_type == "int":
            int_columns.append(c_name)

        elif c_type == "numeric":
            num_columns.append(c_name)

    # Converting all observations to string
    obs_df = table_df[str_columns + int_columns]
    obs_df.index = table_df.index

    x_df = table_df[num_columns]

    if x_df.dtypes.nunique() > 1:
        x_df = x_df.astype("float64")

    if x_df.empty:
        # If there are no numeric columns, create an empty array
        # to avoid AnnData failing to create the object
        x_df = np.zeros((0, 0), dtype="float64")

    return ad.AnnData(X=x_df, obs=obs_df)


def table_ad_to_df(
    table_ad: ad.AnnData,
    index_key: str = "label",
    index_type: Literal["str", "int"] = "int",
    validators: list[Validator] | None = None,
) -> pd.DataFrame:
    """Convert a AnnData object representing a fractal table to a pandas DataFrame.

    Args:
        table_ad (ad.AnnData): An AnnData object representing a fractal table.
        index_key (str): The column name to use as the index of the DataFrame.
            Default is 'label'.
        index_type (str): The type of the index column in the DataFrame.
            Either 'str' or 'int'. Default is 'int'.
        validators (list[Validator]): A list of functions to further validate the table.
    """
    table_df = table_ad.to_df()

    table_df[table_ad.obs_keys()] = table_ad.obs

    # Set the index of the DataFrame
    if table_ad.obs.index.name is not None:
        table_df.index = table_ad.obs.index
    elif index_key in table_df.columns:
        table_df = table_df.set_index(index_key)
    else:
        raise TableValidationError(
            f"Index key {index_key} not found in AnnData object."
        )

    table_df = validate_table(
        table_df=table_df,
        index_key=index_key,
        index_type=index_type,
        validators=validators,
    )
    return table_df


####################################################################################################
#
# Common table validators
#
####################################################################################################


def validate_columns(
    table_df: pd.DataFrame,
    required_columns: list[str],
    optional_columns: list[str] | None = None,
) -> pd.DataFrame:
    """Validate the columns headers of the table.

    If a required column is missing, a TableValidationError is raised.
    If a list of optional columns is provided, only required and optional columns are
        allowed in the table.

    Args:
        table_df (pd.DataFrame): The DataFrame to validate.
        required_columns (list[str]): A list of required columns.
        optional_columns (list[str] | None): A list of optional columns.
            Default is None.

    Returns:
        pd.DataFrame: The validated DataFrame.
    """
    table_header = table_df.columns
    for column in required_columns:
        if column not in table_header:
            raise TableValidationError(f"Column {column} is required in ROI table")

    if optional_columns is None:
        return table_df

    possible_columns = [*required_columns, *optional_columns]
    for column in table_header:
        if column not in possible_columns:
            raise TableValidationError(
                f"Column {column} is not recognized in ROI table"
            )

    return table_df


def validate_unique_index(table_df: pd.DataFrame) -> pd.DataFrame:
    """Validate that the index of the table is unique."""
    if table_df.index.is_unique:
        return table_df

    # Find the duplicates
    duplicates = table_df.index[table_df.index.duplicated()].tolist()
    raise TableValidationError(
        f"Index of the table contains duplicates values. Duplicate: {duplicates}"
    )



from typing import Protocol  # noqa: I001

import dask.delayed
import numpy as np
from dask import array as da
import dask
from dask.delayed import Delayed

from ngio._common_types import ArrayLike
from ngio.core.roi import RasterCooROI
import zarr


class SlicerTransform(Protocol):
    """A special class of transform that load a specific slice of the data."""

    def get(self, data: ArrayLike) -> ArrayLike:
        """Select a slice of the data and return the result."""
        ...

    def set(
        self,
        data: ArrayLike,
        patch: ArrayLike,
    ) -> None:
        """Replace the slice of the data with the patch and return the result."""
        ...


@dask.delayed
def _slice_set_delayed(
    data: zarr.Array,
    patch: Delayed,
    slices: tuple[slice, ...],
    axes_order: list[int] | None,
) -> None:
    if axes_order is not None:
        patch = da.transpose(patch, axes_order)

    if isinstance(patch, Delayed):
        shape = tuple([s.stop - s.start for s in slices])
        patch = da.from_delayed(patch, shape=shape, dtype=data.dtype)
    da.to_zarr(arr=patch, url=data, region=slices)


class NaiveSlicer:
    """A simple slicer that requires all axes to be specified."""

    def __init__(
        self,
        on_disk_axes_name: list[str],
        axes_order: list[int],
        t: int | slice | None = None,
        c: int | slice | None = None,
        z: int | slice | None = None,
        y: int | slice | None = None,
        x: int | slice | None = None,
        preserve_dimensions: bool = True,
    ):
        """Initialize the NaiveSlicer object."""
        self.on_disk_axes_name = on_disk_axes_name

        # Check if axes_order is trivial
        if axes_order != list(range(len(axes_order))):
            self.axes_order = axes_order
        else:
            self.axes_order = None

        self.slices = {
            "t": self._parse_input(t, preserve_dimensions),
            "c": self._parse_input(c, preserve_dimensions),
            "z": self._parse_input(z, preserve_dimensions),
            "y": self._parse_input(y, preserve_dimensions),
            "x": self._parse_input(x, preserve_dimensions),
        }

        self.slice_on_disk_order = tuple(
            [self.slices[axis] for axis in self.on_disk_axes_name]
        )

    def __repr__(self) -> str:
        """Return the string representation of the object."""
        slices = ", ".join([f"{axis}={slice_}" for axis, slice_ in self.slices.items()])
        return f"NaiveSlicer({slices})"

    def _parse_input(
        self, x: int | slice | None, preserve_dimensions: bool = True
    ) -> slice:
        """Parse the input."""
        if x is None:
            return slice(None)
        elif isinstance(x, int):
            if preserve_dimensions:
                return slice(x, x + 1)
            else:
                return x
        elif isinstance(x, slice):
            return x

        raise ValueError(f"Invalid slice definition {x} of type {type(x)}")

    def _shape_from_slices(self) -> tuple[int, ...]:
        """Return the shape of the slice."""
        slices = self.slice_on_disk_order
        return tuple([s.stop - s.start for s in slices])

    def get(self, data: ArrayLike) -> ArrayLike:
        """Select a slice of the data and return the result."""
        patch = data[self.slice_on_disk_order]

        # If sel.axis_order is trivial, skip the transpose
        if self.axes_order is None:
            return patch

        if isinstance(patch, np.ndarray):
            patch = np.transpose(patch, self.axes_order)
        elif isinstance(patch, da.core.Array):
            patch = da.transpose(patch, self.axes_order)
        else:
            raise ValueError(
                f"Invalid patch type {type(patch)}, "
                "supported types are np.ndarray and da.core.Array"
            )
        return patch

    def set(self, data: ArrayLike, patch: ArrayLike) -> None:
        """Replace the slice of the data with the patch and return the result."""
        # If sel.axis_order is trivial, skip the transpose
        if isinstance(patch, np.ndarray):
            if self.axes_order is not None:
                patch = np.transpose(patch, self.axes_order)
            data[self.slice_on_disk_order] = patch
        elif isinstance(patch, (da.core.Array, Delayed)):  # noqa: UP038
            if self.axes_order is not None:
                patch = da.transpose(patch, self.axes_order)

            if isinstance(patch, Delayed):
                patch = da.from_delayed(
                    patch, shape=self._shape_from_slices(), dtype=data.dtype
                )
            da.to_zarr(arr=patch, url=data, region=self.slice_on_disk_order)
        else:
            raise ValueError(
                f"Invalid patch type {type(patch)}, "
                "supported types are np.ndarray and da.core.Array"
            )


class RoiSlicer(NaiveSlicer):
    """A slicer that requires all axes to be specified."""

    def __init__(
        self,
        on_disk_axes_name: list[str],
        axes_order: list[int],
        roi: RasterCooROI,
        t: int | slice | None = None,
        c: int | slice | None = None,
        preserve_dimensions: bool = True,
    ):
        """Initialize the RoiSlicer object."""
        super().__init__(
            on_disk_axes_name=on_disk_axes_name,
            axes_order=axes_order,
            t=t,
            c=c,
            z=roi.z_slice(),
            y=roi.y_slice(),
            x=roi.x_slice(),
            preserve_dimensions=preserve_dimensions,
        )

    def __repr__(self) -> str:
        """Return the string representation of the object."""
        slices = ", ".join([f"{axis}={slice_}" for axis, slice_ in self.slices.items()])
        return f"RoiSlicer({slices})"



"""Module for handling tables in the Fractal format."""

from ngio.tables.tables_group import (
    FeatureTable,
    MaskingROITable,
    ROITable,
    Table,
    TableGroup,
)

__all__ = ["Table", "ROITable", "FeatureTable", "MaskingROITable", "TableGroup"]



"""Region of interest (ROI) metadata.

These are the interfaces bwteen the ROI tables / masking ROI tables and
    the ImageLikeHandler.
"""

from typing import Any

import numpy as np
from pydantic import BaseModel, Field

from ngio.core.dimensions import Dimensions
from ngio.ngff_meta.fractal_image_meta import PixelSize, SpaceUnits


def _to_raster(value: float, pixel_size: PixelSize, max_shape: int) -> int:
    """Convert to raster coordinates."""
    round_value = int(np.round(value / pixel_size))
    return min(round_value, max_shape)


def _to_world(value: int, pixel_size: float) -> float:
    """Convert to world coordinates."""
    return value * pixel_size


class WorldCooROI(BaseModel):
    """Region of interest (ROI) metadata."""

    x_length: float
    y_length: float
    z_length: float = 1.0
    x: float = 0.0
    y: float = 0.0
    z: float = 0.0
    unit: SpaceUnits = Field(SpaceUnits.micrometer, repr=False)
    infos: dict[str, Any] = Field(default_factory=dict, repr=False)

    def to_raster_coo(
        self, pixel_size: PixelSize, dimensions: Dimensions
    ) -> "RasterCooROI":
        """Convert to raster coordinates."""
        return RasterCooROI(
            x=_to_raster(self.x, pixel_size.x, dimensions.x),
            y=_to_raster(self.y, pixel_size.y, dimensions.y),
            z=_to_raster(self.z, pixel_size.z, dimensions.z),
            x_length=_to_raster(self.x_length, pixel_size.x, dimensions.x),
            y_length=_to_raster(self.y_length, pixel_size.y, dimensions.y),
            z_length=_to_raster(self.z_length, pixel_size.z, dimensions.z),
            original_roi=self,
        )


class RasterCooROI(BaseModel):
    """Region of interest (ROI) metadata."""

    x: int
    y: int
    z: int
    x_length: int
    y_length: int
    z_length: int
    original_roi: WorldCooROI = Field(..., repr=False)

    def to_world_coo_roi(self, pixel_size: PixelSize) -> WorldCooROI:
        """Convert to world coordinates."""
        if self.field_index is None:
            raise ValueError(
                "Field index must be provided to convert to world coordinates roi."
            )
        return WorldCooROI(
            x=_to_world(self.x, pixel_size.x),
            y=_to_world(self.y, pixel_size.y),
            z=_to_world(self.z, pixel_size.z),
            x_length=_to_world(self.x_length, pixel_size.x),
            y_length=_to_world(self.y_length, pixel_size.y),
            z_length=_to_world(self.z_length, pixel_size.z),
            unit=pixel_size.unit,
            infos=self.original_roi.infos,
        )

    def x_slice(self) -> slice:
        """Return the slice for the x-axis."""
        return slice(self.x, self.x + self.x_length)

    def y_slice(self) -> slice:
        """Return the slice for the y-axis."""
        return slice(self.y, self.y + self.y_length)

    def z_slice(self) -> slice:
        """Return the slice for the z-axis."""
        return slice(self.z, self.z + self.z_length)



"""Collection of helper functions to work with Zarr groups."""

from zarr import Group

from ngio.io._zarr import AccessModeLiteral, StoreLike, StoreOrGroup
from ngio.io._zarr_group_utils import (
    open_group_wrapper,
)

# Zarr V3 imports
# from zarr.store.common import StoreLike

__all__ = [
    "Group",
    "StoreLike",
    "AccessModeLiteral",
    "StoreOrGroup",
    "open_group_wrapper",
]



"""Implementation of the FeatureTableV1 class.

This class follows the feature_table specification at:
https://fractal-analytics-platform.github.io/fractal-tasks-core/tables/
"""

from typing import Literal

import pandas as pd
import zarr
from pydantic import BaseModel

from ngio.tables.v1._generic_table import BaseTable, write_table_ad


class FeatureTableV1Meta(BaseModel):
    """Metadata for the ROI table."""

    region: dict[Literal["path"], str] | None = None
    instance_key: str = "label"
    fractal_table_version: Literal["1"] = "1"
    type: Literal["feature_table"] = "feature_table"


class FeatureTableV1:
    """Class to handle fractal Feature tables.

    To know more about the Feature table format, please refer to the
    https://fractal-analytics-platform.github.io/fractal-tasks-core/tables/
    """

    def __init__(
        self,
        group: zarr.Group,
        validate_metadata: bool = True,
        validate_table: bool = True,
        index_key: str | None = None,
    ):
        """Initialize the class from an existing group.

        Args:
            group (zarr.Group): The group containing the
                ROI table.
            validate_metadata (bool): If True, the metadata is validated.
            validate_table (bool): If True, the table is validated.
            index_key (str): The column name to use as the index of the DataFrame.
        """
        if validate_metadata:
            self._meta = FeatureTableV1Meta(**group.attrs)
        else:
            self._meta = FeatureTableV1Meta.model_construct(**group.attrs)

        # Validate the table is not implemented for the Feature table
        validators = None
        validators = validators if validate_table else None

        index_key = index_key if index_key is not None else self._meta.instance_key
        self._table_handler = BaseTable(
            group=group,
            index_key=index_key,
            index_type="int",
            validators=validators,
        )

    @classmethod
    def _new(
        cls,
        parent_group: zarr.Group,
        name: str,
        label_image: str | None = None,
        instance_key: str = "label",
        overwrite: bool = False,
    ) -> None:
        """Create a new Feature table.

        Args:
            parent_group (zarr.Group): The parent group where the ROI table
                will be created.
            name (str): The name of the ROI table.
            label_image (str | None): The relative path to the label image used to
                create the feature table.
            instance_key (str): The column name to use as the index of the DataFrame.
                Default is 'label'.
            overwrite (bool): If True, the table will be overwritten if it exists.
                Default is False.
        """
        group = parent_group.create_group(name, overwrite=overwrite)
        table = pd.DataFrame(index=pd.Index([], name=instance_key, dtype="int"))

        label_image = label_image if label_image is not None else "None"
        meta = FeatureTableV1Meta(
            region={"path": label_image}, instance_key=instance_key
        )
        write_table_ad(
            group=group,
            table=table,
            index_key=instance_key,
            index_type="int",
            meta=meta,
        )
        return cls(group=group)

    @property
    def meta(self) -> FeatureTableV1Meta:
        """Return the metadata of the feature table."""
        return self._meta

    @property
    def table_handler(self) -> BaseTable:
        """Return the table handler."""
        return self._table_handler

    @property
    def table(self) -> pd.DataFrame:
        """Return the feature table as a DataFrame."""
        return self._table_handler.table

    @table.setter
    def table(self, table: pd.DataFrame):
        """Set the feature table."""
        raise NotImplementedError(
            "Setting the table is not implemented. Please use the 'set_table' method."
        )

    def set_table(self, table: pd.DataFrame):
        """Set the feature table."""
        self._table_handler.set_table(table)

    def label_image_name(self, get_full_path: bool = False) -> str:
        """Return the name of the label image.

        The name is assumed to be after the last '/' in the path.
        Since this might fails, get_full_path is True, the full path is returned.

        Args:
            get_full_path (bool): If True, the full path is returned.
        """
        path = self.meta.region["path"]
        if get_full_path:
            return path

        return path.split("/")[-1]

    def write(self):
        """Write the table to the group."""
        self._table_handler.write(meta=self.meta)



import numpy as np


class TestOMEZarrHandlerV04:
    def test_basic_workflow(self, ome_zarr_image_v04_path):
        from ngio.ngff_meta import get_ngff_image_meta_handler

        handler = get_ngff_image_meta_handler(
            store=ome_zarr_image_v04_path, meta_mode="image"
        )

        fractal_meta = handler.load_meta()
        np.testing.assert_array_equal(
            fractal_meta.pixel_size(idx=0).zyx,
            [1.0, 0.1625, 0.1625],
        )
        np.testing.assert_array_equal(
            fractal_meta.scale(idx=0), [1.0, 1.0, 0.1625, 0.1625]
        )
        assert fractal_meta.num_levels == 5
        assert fractal_meta.levels_paths == ["0", "1", "2", "3", "4"]
        assert fractal_meta.channel_labels == ["DAPI", "nanog", "Lamin B1"]
        assert fractal_meta.get_channel_idx(label="DAPI") == 0
        assert fractal_meta.get_channel_idx(wavelength_id="A01_C01") == 0
        assert fractal_meta.axes_names == ["c", "z", "y", "x"]
        assert fractal_meta.space_axes_names == ["z", "y", "x"]
        assert fractal_meta.get_highest_resolution_dataset().path == "0"

    def test_modify_axis_from_metadata(self, ome_zarr_image_v04_path):
        from ngio.ngff_meta import get_ngff_image_meta_handler
        from ngio.ngff_meta.utils import add_axis_to_metadata, remove_axis_from_metadata

        handler = get_ngff_image_meta_handler(
            store=ome_zarr_image_v04_path, meta_mode="image"
        )

        fractal_meta = handler.load_meta()
        meta_no_channel = remove_axis_from_metadata(
            metadata=fractal_meta, axis_name="c"
        )
        assert meta_no_channel.axes_names == ["z", "y", "x"]

        meta_add_channel = add_axis_to_metadata(
            metadata=meta_no_channel, axis_name="c", scale=1.0
        )
        assert meta_add_channel.axes_names == fractal_meta.axes_names

    def test_pixel_size(self, ome_zarr_image_v04_path):
        from ngio.ngff_meta import get_ngff_image_meta_handler

        handler = get_ngff_image_meta_handler(
            store=ome_zarr_image_v04_path, meta_mode="image"
        )

        pixel_size = handler.load_meta().pixel_size(idx=0)
        assert pixel_size.zyx == (1.0, 0.1625, 0.1625)

    def test_modify_axis_from_label_metadata(self, ome_zarr_label_v04_path):
        from ngio.ngff_meta import get_ngff_image_meta_handler

        handler = get_ngff_image_meta_handler(
            store=ome_zarr_label_v04_path, meta_mode="label"
        )

        fractal_meta = handler.load_meta()

        meta_no_channel = fractal_meta.remove_axis(axis_name="z")
        assert meta_no_channel.axes_names == ["y", "x"]

        meta_add_channel = meta_no_channel.add_axis(axis_name="z", scale=1.0)
        assert meta_add_channel.axes_names == fractal_meta.axes_names



"""This module contains the objects to handle the fractal tables V1."""

from ngio.tables.v1.feature_tables import FeatureTableV1
from ngio.tables.v1.masking_roi_tables import MaskingROITableV1
from ngio.tables.v1.roi_tables import ROITableV1

__all__ = ["ROITableV1", "FeatureTableV1", "MaskingROITableV1"]



"""Core classes for the ngio library."""

from ngio.core.image_handler import Image
from ngio.core.label_handler import Label
from ngio.core.ngff_image import NgffImage

__all__ = ["NgffImage", "Image", "Label"]



"""Module for handling the /tables group in an OME-NGFF file.

The /tables group contains t
"""

from typing import Literal
from warnings import warn

import zarr
from pydantic import ValidationError

from ngio.io import StoreLike
from ngio.pydantic_utils import BaseWithExtraFields
from ngio.tables.v1 import FeatureTableV1, MaskingROITableV1, ROITableV1

ROITable = ROITableV1
IMPLEMENTED_ROI_TABLES = {"1": ROITableV1}

FeatureTable = FeatureTableV1
IMPLEMENTED_FEATURE_TABLES = {"1": FeatureTableV1}

MaskingROITable = MaskingROITableV1
IMPLEMENTED_MASKING_ROI_TABLES = {"1": MaskingROITableV1}

Table = ROITable | FeatureTable | MaskingROITable

TableType = Literal["roi_table", "feature_table", "masking_roi_table"]


class CommonMeta(BaseWithExtraFields):
    """Common metadata for all tables."""

    type: Literal["roi_table", "feature_table", "masking_roi_table"]
    fractal_table_version: str = "1"


def _find_table_impl(
    table_type: Literal["roi_table", "feature_table", "masking_roi_table"],
    version: str,
) -> Table:
    """Find the type of table in the group."""
    if table_type == "roi_table":
        if version not in IMPLEMENTED_ROI_TABLES:
            raise ValueError(f"ROI Table version {version} not implemented.")
        return IMPLEMENTED_ROI_TABLES[version]

    elif table_type == "feature_table":
        if version not in IMPLEMENTED_FEATURE_TABLES:
            raise ValueError(f"Feature Table version {version} not implemented.")
        return IMPLEMENTED_FEATURE_TABLES[version]

    elif table_type == "masking_roi_table":
        if version not in IMPLEMENTED_MASKING_ROI_TABLES:
            raise ValueError(f"Masking ROI Table version {version} not implemented.")
        return IMPLEMENTED_MASKING_ROI_TABLES[version]

    else:
        raise ValueError(f"Table type {table_type} not recognized.")


def _get_table_impl(
    group: zarr.Group,
    validate_metadata: bool = True,
    table_type: TableType | None = None,
    validate_table: bool = True,
    index_key: str | None = None,
) -> Table:
    if validate_metadata:
        common_meta = CommonMeta(**group.attrs)
        table_type = common_meta.type
    else:
        common_meta = CommonMeta.model_construct(**group.attrs)
        if table_type is None:
            raise ValueError(
                "Table type must be provided if metadata is not validated."
            )

    version = common_meta.fractal_table_version
    return _find_table_impl(table_type=table_type, version=version)(
        group=group,
        validate_metadata=validate_metadata,
        validate_table=validate_table,
        index_key=index_key,
    )


class TableGroup:
    """A class to handle the /labels group in an OME-NGFF file."""

    def __init__(self, group: StoreLike | zarr.Group) -> None:
        """Initialize the LabelGroupHandler."""
        if not isinstance(group, zarr.Group):
            group = zarr.open_group(group, mode="a")

        if "tables" not in group:
            self._group = group.create_group("tables")
        else:
            self._group: zarr.Group = group["tables"]

    def _validate_list_of_tables(self, list_of_tables: list[str]) -> None:
        """Validate the list of tables."""
        list_of_groups = list(self._group.group_keys())

        for table_name in list_of_tables:
            if table_name not in list_of_groups:
                raise ValueError(f"Table {table_name} not found in the group.")

    def _get_list_of_tables(self) -> list[str]:
        """Return the list of tables."""
        list_of_tables = self._group.attrs.get("tables", [])
        self._validate_list_of_tables(list_of_tables)
        return list_of_tables

    def list(
        self,
        table_type: TableType | None = None,
    ) -> list[str]:
        """List all labels in the group.

        Args:
            table_type (str): The type of table to list.
                If None, all tables are listed.
                Allowed values are: 'roi_table', 'feature_table', 'masking_roi_table'.
        """
        list_of_tables = self._get_list_of_tables()
        self._validate_list_of_tables(list_of_tables=list_of_tables)
        if table_type is None:
            return list_of_tables

        else:
            if table_type not in ["roi_table", "feature_table", "masking_roi_table"]:
                raise ValueError(
                    f"Table type {table_type} not recognized. "
                    " Allowed values are: 'roi', 'feature', 'masking_roi'."
                )
            list_of_typed_tables = []
            for table_name in list_of_tables:
                table = self._group[table_name]
                try:
                    common_meta = CommonMeta(**table.attrs)
                    if common_meta.type == table_type:
                        list_of_typed_tables.append(table_name)
                except ValidationError:
                    warn(
                        f"Table {table_name} metadata is not correctly formatted.",
                        stacklevel=1,
                    )
            return list_of_typed_tables

    def get_table(
        self,
        name: str,
        table_type: TableType | None = None,
        validate_metadata: bool = True,
        validate_table: bool = True,
        index_key: str | None = None,
    ) -> Table:
        """Get a label from the group.

        Args:
            name (str): The name of the table to get.
            table_type (str): The type of table to get.
                If None, all the table type will be inferred from the metadata.
                Allowed values are: 'roi_table', 'feature_table', 'masking_roi_table'.
            validate_metadata (bool): Whether to validate the metadata of the table.
            validate_table (bool): Whether to validate the table.
            index_key (str): The column name to use as the index of the DataFrame.
                This is usually defined in the metadata of the table, if given here,
                it will overwrite the metadata.
        """
        list_of_tables = self._get_list_of_tables()
        if name not in list_of_tables:
            raise ValueError(f"Table {name} not found in the group.")

        return _get_table_impl(
            group=self._group[name],
            validate_metadata=validate_metadata,
            table_type=table_type,
            validate_table=validate_table,
            index_key=index_key,
        )

    def new(
        self,
        name: str,
        table_type: str = "roi_table",
        overwrite: bool = False,
        version: str = "1",
        **type_specific_kwargs,
    ) -> Table:
        """Add a new table to the group."""
        list_of_tables = self._get_list_of_tables()
        if not overwrite and name in list_of_tables:
            raise ValueError(f"Table {name} already exists in the group.")

        if overwrite and name in list_of_tables:
            list_of_tables.remove(name)

        table_impl = _find_table_impl(table_type=table_type, version=version)
        new_table = table_impl._new(
            parent_group=self._group,
            name=name,
            overwrite=overwrite,
            **type_specific_kwargs,
        )

        self._group.attrs["tables"] = [*list_of_tables, name]

        return new_table



from pathlib import Path

import zarr

from ngio.io._zarr import (
    AccessModeLiteral,
    StoreLike,
    StoreOrGroup,
    ZarrFormat,
    _open_group_v2_v3,
    _pass_through_group,
)

# Zarr v3 Imports
# import zarr.store
# from zarr.core.common import AccessModeLiteral, ZarrFormat
# from zarr.store.common import StoreLike


def _check_store(store: StoreLike) -> StoreLike:
    if isinstance(store, str) or isinstance(store, Path):
        return store

    raise NotImplementedError(
        "RemoteStore is not yet supported. Please use LocalStore."
    )


def open_group_wrapper(
    store: StoreOrGroup, mode: AccessModeLiteral, zarr_format: ZarrFormat = 2
) -> zarr.Group:
    """Wrapper around zarr.open_group with some additional checks.

    Args:
        store (StoreOrGroup): The store (can also be a Path/str) or group to open.
        mode (ReadOrEdirLiteral): The mode to open the group in.
        zarr_format (ZarrFormat): The Zarr format to use.

    Returns:
        zarr.Group: The opened Zarr group.
    """
    if isinstance(store, zarr.Group):
        return _pass_through_group(store, mode=mode, zarr_format=zarr_format)

    store = _check_store(store)

    return _open_group_v2_v3(store=store, mode=mode, zarr_format=zarr_format)



"""Implementation of a Generic Table class."""

from pathlib import Path
from typing import Literal

import anndata as ad
import pandas as pd
import zarr
from pydantic import BaseModel

from ngio.tables._utils import Validator, table_ad_to_df, table_df_to_ad, validate_table

REQUIRED_COLUMNS = [
    "x_micrometer",
    "y_micrometer",
    "z_micrometer",
    "len_x_micrometer",
    "len_y_micrometer",
    "len_z_micrometer",
]


def write_table_ad(
    group: zarr.Group,
    table: pd.DataFrame,
    index_key: str,
    index_type: Literal["int", "str"],
    meta: BaseModel,
    validators: list[Validator] | None = None,
) -> None:
    """Write a table to a Zarr group.

    Args:
        group (zarr.Group): The group to write the table to.
        table (pd.DataFrame): The table to write.
        index_key (str): The column name to use as the index of the DataFrame.
        index_type (str): The type of the index column in the DataFrame.
        meta (BaseModel): The metadata of the table.
        validators (list[Validator]): A list of functions to further validate the
            table.
    """
    ad_table = table_df_to_ad(
        table,
        index_key=index_key,
        index_type=index_type,
        validators=validators,
    )

    group_path = Path(group.store.path) / group.path
    ad_table.write_zarr(group_path)
    group.attrs.update(meta.model_dump(exclude=None))


class BaseTable:
    """A base table class to be used for table operations in all other tables."""

    def __init__(
        self,
        group: zarr.Group,
        index_key: str,
        index_type: Literal["int", "str"],
        validators: list[Validator] | None = None,
    ):
        """Initialize the class from an existing group.

        Args:
            group (zarr.Group): The group containing the
                ROI table.
            index_key (str): The column name to use as the index of the DataFrame.
            index_type (str): The type of the index column in the DataFrame.
            validators (list[Validator]): A list of functions to further validate the
                table.
        """
        self._table_group = group
        self._index_key = index_key
        self._index_type = index_type
        self._validators = validators

        table_ad = ad.read_zarr(self._table_group)

        self._table = table_ad_to_df(
            table_ad=table_ad,
            index_key=self._index_key,
            index_type=self._index_type,
            validators=self._validators,
        )

    @property
    def table(self) -> pd.DataFrame:
        """Return the ROI table as a DataFrame."""
        return self._table

    @table.setter
    def table(self, table: pd.DataFrame) -> None:
        raise NotImplementedError(
            "Setting the table directly is not supported. "
            "Please use the 'set_table' method."
        )

    def set_table(self, table: pd.DataFrame) -> None:
        table = validate_table(
            table_df=table,
            index_key=self.index_key,
            index_type=self.index_type,
            validators=self._validators,
        )
        self._table = table

    def as_anndata(self) -> ad.AnnData:
        """Return the ROI table as an AnnData object."""
        return table_df_to_ad(
            self.table, index_key=self.index_key, index_type=self.index_type
        )

    def from_anndata(self, table_ad: ad.AnnData) -> None:
        """Return the ROI table as an AnnData object."""
        table = table_ad_to_df(
            table_ad=table_ad,
            index_key=self.index_key,
            index_type=self.index_type,
            validators=self._validators,
        )
        # Don't use the setter to avoid re-validating the table
        self._table = table

    @property
    def index(self) -> list[int | str]:
        """Return a list of all the labels in the table."""
        return self.table.index.tolist()

    @property
    def group(self) -> zarr.Group:
        """Return the group of the table."""
        return self._table_group

    @property
    def group_path(self) -> Path:
        """Return the path of the group."""
        return Path(self._table_group.store.path) / self._table_group.path

    @property
    def index_key(self) -> str:
        """Return the index key of the table."""
        return self._index_key

    @property
    def index_type(self) -> Literal["int", "str"]:
        """Return the index type of the table."""
        return self._index_type

    @property
    def validators(self) -> list[Validator] | None:
        """Return the validators of the table."""
        return self._validators

    @validators.setter
    def validators(self, validators: list[Validator] | None) -> None:
        """Set the validators of the table."""
        self._validators = validators

    def add_validator(self, validator: Validator) -> None:
        """Add a validator to the table."""
        if self._validators is None:
            self._validators = []
        self._validators.append(validator)

    def write(self, meta: BaseModel) -> None:
        """Write the current state of the table to the Zarr file."""
        table = self.table
        table = validate_table(
            table_df=table,
            index_key=self.index_key,
            index_type=self.index_type,
            validators=self._validators,
        )
        write_table_ad(
            group=self._table_group,
            table=self.table,
            index_key=self.index_key,
            index_type=self.index_type,
            meta=meta,
            validators=self._validators,
        )



class TestImageHandler:
    def test_ngff_image(self, ome_zarr_image_v04_path):
        import numpy as np

        from ngio.core.image_handler import Image

        image_handler = Image(store=ome_zarr_image_v04_path, path="0")

        assert image_handler.channel_labels == ["DAPI", "nanog", "Lamin B1"]
        assert image_handler.get_channel_idx(label="DAPI") == 0
        assert image_handler.get_channel_idx(wavelength_id="A01_C01") == 0

        shape = image_handler.shape
        assert image_handler.get_array(c=0).shape == shape[1:]
        assert (
            image_handler.get_array(c=0, preserve_dimensions=True).shape
            == (1,) + shape[1:]
        )

        image_handler.set_array(patch=np.ones((3, 10, 256, 256), dtype=np.uint16))
        assert image_handler.get_array(c=0, t=0, z=0, x=0, y=0) == 1

        image_handler.consolidate(order=0)

        image_handler_1 = Image(store=ome_zarr_image_v04_path, path="1")
        assert image_handler_1.get_array(c=0, t=0, z=0, x=0, y=0) == 1



import numpy as np


class TestUtils:
    def test_create_fractal_meta_with_t(self):
        from ngio.ngff_meta import create_image_metadata

        meta = create_image_metadata(
            on_disk_axis=("t", "c", "z", "y", "x"),
            pixel_sizes=None,
            xy_scaling_factor=2.0,
            z_scaling_factor=1.0,
            time_spacing=1.0,
            time_units="s",
            num_levels=5,
            name="test",
            channel_labels=["DAPI", "nanog", "Lamin B1"],
            channel_wavelengths=["A01_C01", "A02_C02", "A03_C03"],
            channel_kwargs=None,
            omero_kwargs=None,
            version="0.4",
        )

        assert meta.channel_labels == ["DAPI", "nanog", "Lamin B1"]
        np.testing.assert_array_equal(meta.pixel_size(idx=0).zyx, [1.0, 1.0, 1.0])
        np.testing.assert_array_equal(meta.scale(idx=0), [1.0, 1.0, 1.0, 1.0, 1.0])
        np.testing.assert_array_equal(meta.pixel_size(path="2").zyx, [1.0, 4.0, 4.0])
        np.testing.assert_array_equal(meta.scale(path="2"), [1.0, 1.0, 1.0, 4.0, 4.0])

        assert meta.num_levels == 5

    def test_create_fractal_meta(self):
        from ngio.ngff_meta import create_image_metadata

        meta = create_image_metadata(
            on_disk_axis=("c", "z", "y", "x"),
            pixel_sizes=None,
            xy_scaling_factor=2.0,
            z_scaling_factor=1.0,
            time_spacing=1.0,
            time_units="s",
            num_levels=5,
            name="test",
            channel_labels=["DAPI", "nanog", "Lamin B1"],
            channel_wavelengths=["A01_C01", "A02_C02", "A03_C03"],
            channel_kwargs=None,
            omero_kwargs=None,
            version="0.4",
        )

        assert meta.channel_labels == ["DAPI", "nanog", "Lamin B1"]
        np.testing.assert_array_equal(meta.pixel_size(idx=0).zyx, [1.0, 1.0, 1.0])
        np.testing.assert_array_equal(meta.scale(idx=0), [1.0, 1.0, 1.0, 1.0])
        np.testing.assert_array_equal(meta.pixel_size(path="2").zyx, [1.0, 4.0, 4.0])
        np.testing.assert_array_equal(meta.scale(path="2"), [1.0, 1.0, 4.0, 4.0])

        assert meta.num_levels == 5

    def test_create_fractal_meta_with_non_canonical_order(self):
        from ngio.ngff_meta import create_image_metadata

        meta = create_image_metadata(
            on_disk_axis=("z", "c", "y", "x"),
            pixel_sizes=None,
            xy_scaling_factor=2.0,
            z_scaling_factor=1.0,
            time_spacing=1.0,
            time_units="s",
            num_levels=5,
            name="test",
            channel_labels=["DAPI", "nanog", "Lamin B1"],
            channel_wavelengths=["A01_C01", "A02_C02", "A03_C03"],
            channel_kwargs=None,
            omero_kwargs=None,
            version="0.4",
        )

        assert meta.axes_names == ["c", "z", "y", "x"]
        assert meta.space_axes_names == ["z", "y", "x"]

        meta = create_image_metadata(
            on_disk_axis=("t", "c", "z", "y", "x"),
            pixel_sizes=None,
            xy_scaling_factor=2.0,
            z_scaling_factor=1.0,
            time_spacing=1.0,
            time_units="s",
            num_levels=5,
            name="test",
            channel_labels=["DAPI", "nanog", "Lamin B1"],
            channel_wavelengths=["A01_C01", "A02_C02", "A03_C03"],
            channel_kwargs=None,
            omero_kwargs=None,
            version="0.4",
        )

        assert meta.axes_names == ["t", "c", "z", "y", "x"]
        assert meta.space_axes_names == ["z", "y", "x"]

    def test_create_fractal_label_meta(self):
        from ngio.ngff_meta import create_label_metadata

        meta = create_label_metadata(
            on_disk_axis=("t", "z", "y", "x"),
            pixel_sizes=None,
            xy_scaling_factor=2.0,
            z_scaling_factor=1.0,
            time_spacing=1.0,
            time_units="s",
            num_levels=5,
            name="test",
            version="0.4",
        )

        np.testing.assert_array_equal(meta.pixel_size(idx=0).zyx, (1.0, 1.0, 1.0))
        np.testing.assert_array_equal(meta.scale(idx=0), [1.0, 1.0, 1.0, 1.0])
        np.testing.assert_array_equal(meta.pixel_size(path="2").zyx, (1.0, 4.0, 4.0))
        np.testing.assert_array_equal(meta.scale(path="2"), [1.0, 1.0, 4.0, 4.0])

        assert meta.num_levels == 5



"""Zarr utilities for loading metadata from OME-NGFF 0.4."""

from typing import Literal

from ngio.io import (
    AccessModeLiteral,
    Group,
    StoreOrGroup,
    open_group_wrapper,
)
from ngio.io._zarr import _is_group_readonly
from ngio.ngff_meta.fractal_image_meta import (
    Axis,
    Dataset,
    ImageLabelMeta,
    ImageMeta,
    LabelMeta,
    Omero,
)
from ngio.ngff_meta.v04.specs import (
    Axis04,
    Dataset04,
    Multiscale04,
    NgffImageMeta04,
    Omero04,
    ScaleCoordinateTransformation04,
    Transformation04,
    TranslationCoordinateTransformation04,
)


def check_ngff_image_meta_v04(store: StoreOrGroup) -> bool:
    """Check if a Zarr Group contains the OME-NGFF v0.4."""
    store = open_group_wrapper(store=store, mode="r", zarr_format=2)
    attrs = dict(store.attrs)
    multiscales = attrs.get("multiscales", None)
    if multiscales is None:
        return False

    version = multiscales[0].get("version", None)
    if version != "0.4":
        return False

    return True


def load_vanilla_ngff_image_meta_v04(group: Group) -> NgffImageMeta04:
    """Load the OME-NGFF 0.4 image meta model."""
    return NgffImageMeta04(**group.attrs)


def _transform_dataset(
    datasets04: list[Dataset04],
    axes: list[Axis04],
    coo_transformation04: list[Transformation04] | None,
) -> list[Dataset]:
    # coo_transformation validation
    # only one scale transformation is allowed as top-level
    if coo_transformation04 is not None:
        if len(coo_transformation04) != 1:
            raise ValueError("Only one scale transformation is allowed as top-level.")

        if not isinstance(coo_transformation04[0], ScaleCoordinateTransformation04):
            raise ValueError(
                "Invalid coordinate transformations. \
                Only scale transformation is allowed."
            )

        top_scale = coo_transformation04[0].scale
    else:
        top_scale = None

    fractal_datasets = []
    for dataset04 in datasets04:
        scale, translation = None, None
        for transformation in dataset04.coordinateTransformations:
            if isinstance(transformation, TranslationCoordinateTransformation04):
                translation = transformation.translation

            if isinstance(transformation, ScaleCoordinateTransformation04):
                scale = transformation.scale
                if top_scale is not None:
                    # Combine the scale transformation with the top-level scale
                    if len(scale) != len(top_scale):
                        raise ValueError(
                            "Inconsistent scale transformation. \
                            The scale transformation must have the same length."
                        )
                    # Combine the scale transformation with the top-level scale
                    scale = [s * ts for s, ts in zip(scale, top_scale, strict=True)]
                scale = scale
        fractal_datasets.append(
            Dataset(
                path=dataset04.path,
                on_disk_axes=axes,
                on_disk_scale=scale,
                on_disk_translation=translation,
            )
        )
    return fractal_datasets


def vanilla_ngff_image_meta_v04_to_fractal(
    meta04: NgffImageMeta04,
    meta_mode: Literal["image", "label"] = "image",
) -> ImageLabelMeta:
    """Convert the NgffImageMeta04 to ImageMeta."""
    if not isinstance(meta04, NgffImageMeta04):
        raise ValueError("Invalid metadata type. Expected NgffImageMeta04.")

    multiscale04 = meta04.multiscales[0]
    axes = [Axis(name=axis.name, unit=axis.unit) for axis in multiscale04.axes]
    fractal_datasets = _transform_dataset(
        datasets04=multiscale04.datasets,
        axes=axes,
        coo_transformation04=multiscale04.coordinateTransformations,
    )

    if meta_mode == "label":
        return LabelMeta(
            version="0.4",
            name=multiscale04.name,
            datasets=fractal_datasets,
        )

    fractal_omero = None if meta04.omero is None else Omero(**meta04.omero.model_dump())

    return ImageMeta(
        version="0.4",
        name=multiscale04.name,
        datasets=fractal_datasets,
        omero=fractal_omero,
    )


def fractal_ngff_image_meta_to_vanilla_v04(
    meta: ImageLabelMeta,
) -> NgffImageMeta04:
    """Convert the ImageMeta to NgffImageMeta."""
    axes04 = [Axis04(**axis.model_dump()) for axis in meta.axes]
    dataset04 = []
    for dataset in meta.datasets:
        transformations = [
            ScaleCoordinateTransformation04(type="scale", scale=dataset.scale)
        ]
        if dataset.translation is not None:
            transformations.append(
                TranslationCoordinateTransformation04(
                    type="translation", translation=dataset.translation
                )
            )
        dataset04.append(
            Dataset04(path=dataset.path, coordinateTransformations=transformations)
        )
    multiscale04 = Multiscale04(
        name=meta.name,
        axes=axes04,
        datasets=dataset04,
        version="0.4",
    )

    if isinstance(meta, LabelMeta):
        return NgffImageMeta04(multiscales=[multiscale04])

    omero04 = None if meta.omero is None else Omero04(**meta.omero.model_dump())
    return NgffImageMeta04(
        multiscales=[multiscale04],
        omero=omero04,
    )


def load_ngff_image_meta_v04(group: Group) -> ImageLabelMeta:
    """Load the OME-NGFF 0.4 image meta model."""
    if not check_ngff_image_meta_v04(store=group):
        raise ValueError(
            "The Zarr store does not contain the correct OME-Zarr version."
        )
    meta04 = load_vanilla_ngff_image_meta_v04(group=group)
    return vanilla_ngff_image_meta_v04_to_fractal(meta04=meta04)


def write_ngff_image_meta_v04(group: Group, meta: ImageLabelMeta) -> None:
    """Write the OME-NGFF 0.4 image meta model."""
    if dict(group.attrs):
        # If group is not empty, check if the version is correct
        if not check_ngff_image_meta_v04(store=group):
            raise ValueError(
                "The Zarr store does not contain the correct OME-Zarr version."
            )

    if isinstance(meta, ImageMeta) and meta.omero is not None:
        for c in meta.omero.channels:
            if "color" not in c.extra_fields:
                c.extra_fields["color"] = "00FFFF"

    meta04 = fractal_ngff_image_meta_to_vanilla_v04(meta=meta)
    group.attrs.update(meta04.model_dump(exclude=None))


class NgffImageMetaZarrHandlerV04:
    """Class for loading and writing OME-NGFF 0.4 metadata."""

    def __init__(
        self,
        store: StoreOrGroup,
        meta_mode: Literal["image", "label"],
        cache: bool = False,
        mode: AccessModeLiteral = "a",
    ):
        """Initialize the handler.

        Args:
            store (StoreOrGroup): The Zarr store or group containing the image data.
            meta_mode (str): The mode of the metadata handler.
            cache (bool): Whether to cache the metadata.
            mode (str): The mode of the store.
        """
        if isinstance(store, Group):
            if hasattr(store, "store_path"):
                self._store = store.store_path
            else:
                self._store = store.store

            self._group = store

        else:
            self._store = store
            self._group = open_group_wrapper(store=store, mode=mode, zarr_format=2)

        self.meta_mode = meta_mode
        self.cache = cache
        self._meta = None

    @property
    def zarr_version(self) -> int:
        """Return the Zarr version.

        This is not strictly necessary, but it is necessary
        to make sure the zarr python creare consistent zarr files.
        """
        return 2

    @property
    def store(self):
        """Return the Zarr store."""
        return self._store

    @property
    def group(self) -> Group:
        """Return the Zarr group."""
        return self._group

    @staticmethod
    def check_version(store: StoreOrGroup) -> bool:
        """Check if the version of the metadata is supported."""
        return check_ngff_image_meta_v04(store=store)

    def load_meta(self) -> ImageLabelMeta:
        """Load the OME-NGFF 0.4 metadata."""
        if not self.check_version(store=self.group):
            raise ValueError(
                "The Zarr store does not contain the correct OME-Zarr version."
            )

        if self.cache:
            if self._meta is None:
                self._meta = load_ngff_image_meta_v04(self.group)
            return self._meta

        return load_ngff_image_meta_v04(self.group)

    def write_meta(self, meta: ImageLabelMeta) -> None:
        """Write the OME-NGFF 0.4 metadata."""
        if _is_group_readonly(self.group):
            raise ValueError(
                "The store is read-only. Cannot write the metadata to the store."
            )

        write_ngff_image_meta_v04(group=self.group, meta=meta)

        if self.cache:
            self.update_cache(meta)

    def update_cache(self, meta: ImageLabelMeta) -> None:
        """Update the cached metadata."""
        if not self.cache:
            raise ValueError("Cache is not enabled.")
        self._meta = meta

    def clear_cache(self) -> None:
        """Clear the cached metadata."""
        self._meta = None



import json
from importlib.metadata import version
from pathlib import Path

import zarr
from packaging.version import Version
from pytest import fixture

zarr_version = version("zarr")
ZARR_PYTHON_V = 2 if Version(zarr_version) < Version("3.0.0a") else 3


@fixture
def ome_zarr_image_v04_path(tmpdir):
    zarr_path = Path(tmpdir) / "test_ome_ngff_v04.zarr"

    if ZARR_PYTHON_V == 3:
        group = zarr.open_group(store=zarr_path, mode="w", zarr_format=2)
    else:
        group = zarr.open_group(store=zarr_path, mode="w", zarr_version=2)

    with open("tests/data/meta_v04/base_ome_zarr_image_meta.json") as f:
        base_ome_zarr_meta = json.load(f)

    base_ome_zarr_meta = base_ome_zarr_meta
    group.attrs.update(base_ome_zarr_meta)
    return zarr_path


@fixture
def ome_zarr_label_v04_path(tmpdir):
    zarr_path = Path(tmpdir) / "test_ome_ngff_image_v04.zarr"

    if ZARR_PYTHON_V == 3:
        group = zarr.open_group(store=zarr_path, mode="w", zarr_format=2)
    else:
        group = zarr.open_group(store=zarr_path, mode="w", zarr_version=2)

    with open("tests/data/meta_v04/base_ome_zarr_label_meta.json") as f:
        base_ome_zarr_meta = json.load(f)

    base_ome_zarr_meta = base_ome_zarr_meta
    group.attrs.update(base_ome_zarr_meta)
    return zarr_path



"""Abstract class for handling OME-NGFF images."""

import numpy as np

from ngio.core.image_handler import Image
from ngio.core.label_handler import LabelGroup
from ngio.core.utils import create_empty_ome_zarr_image
from ngio.io import StoreLike, open_group_wrapper
from ngio.ngff_meta import get_ngff_image_meta_handler
from ngio.ngff_meta.fractal_image_meta import ImageMeta, PixelSize
from ngio.tables.tables_group import TableGroup


class NgffImage:
    """A class to handle OME-NGFF images."""

    def __init__(self, store: StoreLike, cache: bool = False) -> None:
        """Initialize the NGFFImage in read mode."""
        self.store = store
        self.group = open_group_wrapper(store=store, mode="r+")
        self._image_meta = get_ngff_image_meta_handler(
            self.group, meta_mode="image", cache=cache
        )
        self._metadata_cache = cache
        self.table = TableGroup(self.group)
        self.label = LabelGroup(self.group, image_ref=self.get_image())

    @property
    def image_meta(self) -> ImageMeta:
        """Get the image metadata."""
        return self._image_meta.load_meta()

    @property
    def num_levels(self) -> int:
        """Get the number of levels in the image."""
        return self.image_meta.num_levels

    @property
    def levels_paths(self) -> list[str]:
        """Get the paths of the levels in the image."""
        return self.image_meta.levels_paths

    def get_image(
        self,
        *,
        path: str | None = None,
        pixel_size: PixelSize | None = None,
        highest_resolution: bool = True,
    ) -> Image:
        """Get an image handler for the given level.

        Args:
            path (str | None, optional): The path to the level.
            pixel_size (tuple[float, ...] | list[float] | None, optional): The pixel
                size of the level.
            highest_resolution (bool, optional): Whether to get the highest
                resolution level

        Returns:
            ImageHandler: The image handler.
        """
        if path is not None or pixel_size is not None:
            highest_resolution = False

        return Image(
            store=self.group,
            path=path,
            pixel_size=pixel_size,
            highest_resolution=highest_resolution,
            label_group=LabelGroup(self.group, image_ref=None),
            cache=self._metadata_cache,
        )

    def _update_omero_window(self) -> None:
        """Update the OMERO window."""
        meta = self.image_meta
        image = self.get_image(highest_resolution=True)
        max_dtype = np.iinfo(image.on_disk_array.dtype).max
        start, end = (
            image.on_disk_dask_array.min().compute(),
            image.on_disk_dask_array.max().compute(),
        )

        channel_list = meta.omero.channels

        new_channel_list = []
        for channel in channel_list:
            channel.extra_fields["window"] = {
                "start": start,
                "end": end,
                "min": 0,
                "max": max_dtype,
            }
            new_channel_list.append(channel)

        meta.omero.channels = new_channel_list
        self._image_meta.write_meta(meta)

    def derive_new_image(
        self,
        store: StoreLike,
        name: str,
        overwrite: bool = True,
        **kwargs,
    ) -> "NgffImage":
        """Derive a new image from the current image.

        Args:
            store (StoreLike): The store to create the new image in.
            name (str): The name of the new image.
            overwrite (bool): Whether to overwrite the image if it exists
            **kwargs: Additional keyword arguments.
                Follow the same signature as `create_empty_ome_zarr_image`.

        Returns:
            NgffImage: The new image.
        """
        image_0 = self.get_image(highest_resolution=True)

        default_kwargs = {
            "store": store,
            "shape": image_0.on_disk_shape,
            "chunks": image_0.on_disk_array.chunks,
            "dtype": image_0.on_disk_array.dtype,
            "on_disk_axis": image_0.dataset.on_disk_axes_names,
            "pixel_sizes": image_0.pixel_size,
            "xy_scaling_factor": self.image_meta.xy_scaling_factor,
            "z_scaling_factor": self.image_meta.z_scaling_factor,
            "time_spacing": image_0.dataset.time_spacing,
            "time_units": image_0.dataset.time_axis_unit,
            "num_levels": self.num_levels,
            "name": name,
            "channel_labels": image_0.channel_labels,
            "channel_wavelengths": None,
            "channel_kwargs": None,
            "omero_kwargs": None,
            "overwrite": overwrite,
            "version": self.image_meta.version,
        }

        default_kwargs.update(kwargs)

        create_empty_ome_zarr_image(
            **default_kwargs,
        )
        return NgffImage(store=store)



"""Utility functions for Pydantic models."""

import os
from collections import namedtuple
from typing import TypeVar

from pydantic import BaseModel, Field, model_serializer, model_validator

# Debugging flag to skip validation of the metadata (for testing purposes only)
# check if this is an environment variable
SKIP_NGFF_VALIDATION = bool(os.getenv("SKIP_NGFF_VALIDATION", False))


class BaseWithExtraFields(BaseModel):
    """Base class for all Fractal spec models.

    Every field that is not defined in the model will
        be stored in the `extra_fields` attribute.
    """

    extra_fields: dict = Field(default_factory=dict)

    @model_validator(mode="before")
    def _collect_extra_fields(cls, values):
        extra = {k: v for k, v in values.items() if k not in cls.model_fields}
        values["extra_fields"] = extra
        return values

    @model_serializer(mode="wrap")
    def _custom_serializer(self, handler):
        basic_dict = handler(self)
        extra = basic_dict.pop("extra_fields")
        return {**basic_dict, **extra}


T = TypeVar("T")


def unique_items_validator(values: list[T]) -> list[T]:
    """Validate that all items in the list are unique."""
    if SKIP_NGFF_VALIDATION:
        return values

    for ind, value in enumerate(values, start=1):
        if value in values[ind:]:
            raise ValueError(f"Non-unique values in {values}.")
    return values


def named_tuple_from_pydantic_model(model: BaseModel) -> namedtuple:
    """Create a namedtuple from a Pydantic model."""
    return namedtuple(name=model.__name__, field_names=model.model_fields.keys())



"""Dimension metadata.

This is not related to the NGFF metadata,
but it is based on the actual metadata of the image data.
"""


class Dimensions:
    """Dimension metadata."""

    def __init__(
        self,
        on_disk_shape: tuple[int, ...],
        axes_names: list[str],
        axes_order: list[int],
    ) -> None:
        """Create a Dimension object from a Zarr array.

        Args:
            on_disk_shape (tuple[int, ...]): The shape of the array on disk.
            axes_names (list[str]): The names of the axes in the canonical order.
            axes_order (list[int]): The mapping between the canonical order and the on
                disk order.
        """
        self._on_disk_shape = on_disk_shape

        for s in on_disk_shape:
            if s < 1:
                raise ValueError("The shape must be greater equal to 1.")

        if len(self._on_disk_shape) != len(axes_names):
            raise ValueError(
                "The number of axes names must match the number of dimensions."
            )

        self._axes_names = axes_names
        self._axes_order = axes_order

        self._shape = [self._on_disk_shape[i] for i in axes_order]
        self._shape_dict = dict(zip(axes_names, self._shape, strict=True))

    @property
    def shape(self) -> tuple[int, ...]:
        """Return the shape as a tuple in the canonical order."""
        return tuple(self._shape)

    @property
    def on_disk_shape(self) -> tuple[int, ...]:
        """Return the shape as a tuple."""
        return self._on_disk_shape

    def ad_dict(self) -> dict[str, int]:
        """Return the shape as a dictionary."""
        return self._shape_dict

    @property
    def t(self) -> int:
        """Return the time dimension."""
        return self._shape_dict.get("t", None)

    @property
    def c(self) -> int:
        """Return the channel dimension."""
        return self._shape_dict.get("c", None)

    @property
    def z(self) -> int:
        """Return the z dimension."""
        return self._shape_dict.get("z", None)

    @property
    def y(self) -> int:
        """Return the y dimension."""
        return self._shape_dict.get("y")

    @property
    def x(self) -> int:
        """Return the x dimension."""
        return self._shape_dict.get("x")

    @property
    def on_disk_ndim(self) -> int:
        """Return the number of dimensions on disk."""
        return len(self._on_disk_shape)

    @property
    def is_time_series(self) -> bool:
        """Return whether the data is a time series."""
        if (self.t is None) or (self.t == 1):
            return False
        return True

    @property
    def is_2d(self) -> bool:
        """Return whether the data is 2D."""
        if (self.z is not None) and (self.z > 1):
            return False
        return True

    @property
    def is_2d_time_series(self) -> bool:
        """Return whether the data is a 2D time series."""
        is_2d = self.is_2d()
        is_time_series = self.is_time_series()
        return is_2d and is_time_series

    @property
    def is_3d(self) -> bool:
        """Return whether the data is 3D."""
        if (self.z is None) or (self.z == 1):
            return False
        return True

    @property
    def is_3d_time_series(self) -> bool:
        """Return whether the data is a 3D time series."""
        is_3d = self.is_3d()
        is_time_series = self.is_time_series()
        return is_3d and is_time_series

    @property
    def is_multi_channels(self) -> bool:
        """Return whether the data has multiple channels."""
        if (self.c is None) or (self.c == 1):
            return False
        return True



"""IO and validation of NGFF metadata."""

from ngio.ngff_meta.fractal_image_meta import (
    Dataset,
    ImageLabelMeta,
    ImageMeta,
    LabelMeta,
    PixelSize,
    SpaceUnits,
)
from ngio.ngff_meta.meta_handler import get_ngff_image_meta_handler
from ngio.ngff_meta.utils import (
    add_axis_to_metadata,
    create_image_metadata,
    create_label_metadata,
    derive_image_metadata,
    derive_label_metadata,
    remove_axis_from_metadata,
)

__all__ = [
    "Dataset",
    "ImageMeta",
    "LabelMeta",
    "ImageLabelMeta",
    "PixelSize",
    "SpaceUnits",
    "get_ngff_image_meta_handler",
    "add_axis_to_metadata",
    "create_image_metadata",
    "create_label_metadata",
    "derive_image_metadata",
    "derive_label_metadata",
    "remove_axis_from_metadata",
]



"""Generic class to handle Image-like data in a OME-NGFF file."""

from functools import partial
from pathlib import Path
from typing import Literal
from warnings import warn

import dask.array as da
import numpy as np
import zarr
from dask.delayed import Delayed
from scipy.ndimage import zoom

from ngio._common_types import ArrayLike
from ngio.core.dimensions import Dimensions
from ngio.core.roi import WorldCooROI
from ngio.core.utils import Lock
from ngio.io import StoreOrGroup, open_group_wrapper
from ngio.ngff_meta import (
    Dataset,
    ImageLabelMeta,
    PixelSize,
    SpaceUnits,
    get_ngff_image_meta_handler,
)
from ngio.pipes import DataTransformPipe, NaiveSlicer, RoiSlicer


class ImageLike:
    """A class to handle OME-NGFF images stored in Zarr format.

    This class provides methods to access image data and ROI tables.
    """

    def __init__(
        self,
        store: StoreOrGroup,
        path: str | None = None,
        idx: int | None = None,
        pixel_size: PixelSize | None = None,
        highest_resolution: bool = False,
        strict: bool = True,
        meta_mode: Literal["image", "label"] = "image",
        cache: bool = True,
        _label_group=None,
    ) -> None:
        """Initialize the MultiscaleHandler in read mode.

        Note: Only one of `path`, `idx`, 'pixel_size' or 'highest_resolution'
        should be provided.

        store (StoreOrGroup): The Zarr store or group containing the image data.
        path (str | None): The path to the level.
        idx (int | None): The index of the level.
        pixel_size (PixelSize | None): The pixel size of the level.
        highest_resolution (bool): Whether to get the highest resolution level.
        strict (bool): Whether to raise an error where a pixel size is not found
            to match the requested "pixel_size".
        meta_mode (str): The mode of the metadata handler.
        cache (bool): Whether to cache the metadata.
        _label_group (LabelGroup): The group containing the label data (internal use).
        """
        if not strict:
            warn("Strict mode is not fully supported yet.", UserWarning, stacklevel=2)

        if not isinstance(store, zarr.Group):
            store = open_group_wrapper(store=store, mode="r+")

        self._group = store

        self._metadata_handler = get_ngff_image_meta_handler(
            store=store, meta_mode=meta_mode, cache=cache
        )

        # Find the level / resolution index
        metadata = self._metadata_handler.load_meta()
        dataset = metadata.get_dataset(
            path=path,
            idx=idx,
            pixel_size=pixel_size,
            highest_resolution=highest_resolution,
            strict=strict,
        )

        if pixel_size is not None:
            pixel_size.virtual = True
            self._virtual_pixel_size = pixel_size
        else:
            self._virtual_pixel_size = None

        self._init_dataset(dataset)
        self._dask_lock = None

        self._label_group = _label_group

    def _init_dataset(self, dataset: Dataset):
        """Set the dataset of the image.

        This method is for internal use only.
        """
        self._dataset = dataset

        if self._dataset.path not in self._group.array_keys():
            raise ValueError(f"Dataset {self._dataset.path} not found in the group.")

        self._array = self._group[self.dataset.path]
        self._diminesions = Dimensions(
            on_disk_shape=self._array.shape,
            axes_names=self._dataset.axes_names,
            axes_order=self._dataset.axes_order,
        )

    def _debug_set_new_dataset(
        self,
        new_dataset: Dataset,
    ):
        """Debug method to change the the dataset metadata.

        This methods allow to change dataset after initialization.
        This allow to skip the OME-NGFF metadata validation.
        This method is for testing/debug purposes only.

        DO NOT USE THIS METHOD IN PRODUCTION CODE.
        """
        self._init_dataset(new_dataset)

    # Method to get the metadata of the image
    @property
    def group(self) -> zarr.Group:
        """Return the Zarr group containing the image data."""
        return self._group

    @property
    def metadata(self) -> ImageLabelMeta:
        """Return the metadata of the image."""
        return self._metadata_handler.load_meta()

    @property
    def dataset(self) -> Dataset:
        """Return the dataset of the image."""
        return self._dataset

    @property
    def path(self) -> str:
        """Return the path of the dataset (relative to the root group)."""
        return self.dataset.path

    @property
    def channel_labels(self) -> list[str]:
        """Return the names of the channels in the image."""
        return self.dataset.path

    @property
    def axes_names(self) -> list[str]:
        """Return the names of the axes in the image."""
        return self.dataset.axes_names

    @property
    def space_axes_names(self) -> list[str]:
        """Return the names of the space axes in the image."""
        return self.dataset.space_axes_names

    @property
    def space_axes_unit(self) -> SpaceUnits:
        """Return the units of the space axes in the image."""
        return self.dataset.space_axes_unit

    @property
    def pixel_size(self) -> PixelSize:
        """Return the pixel resolution of the image."""
        if self._virtual_pixel_size is not None:
            return self._virtual_pixel_size

        return self.dataset.pixel_size

    # Utility methods to get the image dimensionality
    @property
    def dimensions(self) -> Dimensions:
        """Return an object representation the dimensions of the image."""
        return self._diminesions

    @property
    def shape(self) -> tuple[int, ...]:
        """Return the shape of the image in canonical order (TCZYX)."""
        return self.dimensions.shape

    @property
    def is_time_series(self) -> bool:
        """Return whether the image is a time series."""
        return self.dimensions.is_time_series

    @property
    def is_2d(self) -> bool:
        """Return whether the image is 2D."""
        return self.dimensions.is_2d

    @property
    def is_2d_time_series(self) -> bool:
        """Return whether the image is a 2D time series."""
        return self.dimensions.is_2d_time_series

    @property
    def is_3d(self) -> bool:
        """Return whether the image is 3D."""
        return self.dimensions.is_3d

    @property
    def is_3d_time_series(self) -> bool:
        """Return whether the image is a 3D time series."""
        return self.dimensions.is_3d_time_series

    @property
    def is_multi_channels(self) -> bool:
        """Return whether the image has multiple channels."""
        return self.dimensions.is_multi_channels

    # Methods to get the image data as is on disk
    @property
    def on_disk_array(self) -> zarr.Array:
        """Return the image data as a Zarr array."""
        return self._array

    @property
    def on_disk_dask_array(self) -> da.core.Array:
        """Return the image data as a Dask array."""
        return da.from_zarr(self.on_disk_array)

    @property
    def on_disk_shape(self) -> tuple[int, ...]:
        """Return the shape of the image."""
        return self.dimensions.on_disk_shape

    # Methods to get the image data in the canonical order
    def init_lock(self, lock_id: str | None = None) -> None:
        """Set the lock for the Dask array."""
        if Lock is None:
            raise ImportError(
                "Lock is not available. Please install dask[distributed]."
            )
        # Unique zarr array identifier
        array_path = (
            Path(self._group.store.path) / self._group.path / self._dataset.path
        )
        lock_id = f"Zarr_IO_Lock_{array_path}" if lock_id is None else lock_id
        self._dask_lock = Lock(lock_id)

    def _get_pipe(
        self,
        data_pipe: DataTransformPipe,
        mode: Literal["numpy", "dask"] = "numpy",
    ) -> ArrayLike:
        """Return the data transform pipe."""
        if mode == "numpy":
            return data_pipe.get(data=self.on_disk_array)
        elif mode == "dask":
            return data_pipe.get(data=self.on_disk_dask_array)
        else:
            raise ValueError(f"Invalid mode {mode}")

    def _set_pipe(
        self,
        data_pipe: DataTransformPipe,
        patch: ArrayLike,
    ) -> None:
        """Set the data transform pipe."""
        if isinstance(patch, np.ndarray):
            data_pipe.set(data=self.on_disk_array, patch=patch)

        elif isinstance(patch, (da.core.Array, Delayed)):  # noqa: UP038
            if self._dask_lock is None:
                return data_pipe.set(data=self.on_disk_array, patch=patch)

            array = self.on_disk_array
            with self._dask_lock:
                data_pipe.set(data=array, patch=patch)
        else:
            raise ValueError(
                f"Invalid patch type {type(patch)}. "
                "Supported types are np.ndarray and da.core.Array"
            )

    def _build_roi_pipe(
        self,
        roi: WorldCooROI,
        t: int | slice | None = None,
        c: int | slice | None = None,
        preserve_dimensions: bool = False,
    ) -> DataTransformPipe:
        """Build the data transform pipe for a region of interest (ROI)."""
        roi_coo = roi.to_raster_coo(
            pixel_size=self.dataset.pixel_size, dimensions=self.dimensions
        )
        slicer = RoiSlicer(
            on_disk_axes_name=self.dataset.on_disk_axes_names,
            axes_order=self.dataset.axes_order,
            roi=roi_coo,
            t=t,
            c=c,
            preserve_dimensions=preserve_dimensions,
        )
        return DataTransformPipe(slicer=slicer)

    def _build_naive_pipe(
        self,
        x: int | slice | None = None,
        y: int | slice | None = None,
        z: int | slice | None = None,
        t: int | slice | None = None,
        c: int | slice | None = None,
        preserve_dimensions: bool = False,
    ) -> DataTransformPipe:
        """Build the data transform pipe for a naive slice."""
        slicer = NaiveSlicer(
            on_disk_axes_name=self.dataset.on_disk_axes_names,
            axes_order=self.dataset.axes_order,
            x=x,
            y=y,
            z=z,
            t=t,
            c=c,
            preserve_dimensions=preserve_dimensions,
        )
        return DataTransformPipe(slicer=slicer)

    def get_array_from_roi(
        self,
        roi: WorldCooROI,
        t: int | slice | None = None,
        c: int | slice | None = None,
        mode: Literal["numpy", "dask"] = "numpy",
        preserve_dimensions: bool = False,
    ) -> ArrayLike | tuple[ArrayLike, DataTransformPipe]:
        """Return the image data from a region of interest (ROI).

        Args:
            roi (WorldCooROI): The region of interest.
            t (int | slice | None): The time index or slice.
            c (int | slice | None): The channel index or slice.
            mode (str): The mode to return the data.
            preserve_dimensions (bool): Whether to preserve the dimensions of the data.
        """
        data_pipe = self._build_roi_pipe(
            roi=roi, t=t, c=c, preserve_dimensions=preserve_dimensions
        )
        return_array = self._get_pipe(data_pipe=data_pipe, mode=mode)
        return return_array

    def set_array_from_roi(
        self,
        roi: WorldCooROI,
        patch: ArrayLike,
        t: int | slice | None = None,
        c: int | slice | None = None,
        preserve_dimensions: bool = False,
    ) -> None:
        """Set the image data from a region of interest (ROI).

        Args:
            roi (WorldCooROI): The region of interest.
            patch (ArrayLike): The patch to set.
            t (int | slice | None): The time index or slice.
            c (int | slice | None): The channel index or slice.
            preserve_dimensions (bool): Whether to preserve the dimensions of the data.
        """
        data_pipe = self._build_roi_pipe(
            roi=roi, t=t, c=c, preserve_dimensions=preserve_dimensions
        )
        self._set_pipe(data_pipe=data_pipe, patch=patch)

    def get_array(
        self,
        x: int | slice | None = None,
        y: int | slice | None = None,
        z: int | slice | None = None,
        t: int | slice | None = None,
        c: int | slice | None = None,
        mode: Literal["numpy", "dask"] = "numpy",
        preserve_dimensions: bool = False,
    ) -> ArrayLike:
        """Return the image data.

        Args:
            x (int | slice | None): The x index or slice.
            y (int | slice | None): The y index or slice.
            z (int | slice | None): The z index or slice.
            t (int | slice | None): The time index or slice.
            c (int | slice | None): The channel index or slice.
            mode (str): The mode to return the data.
            preserve_dimensions (bool): Whether to preserve the dimensions of the data.
        """
        data_pipe = self._build_naive_pipe(
            x=x, y=y, z=z, t=t, c=c, preserve_dimensions=preserve_dimensions
        )
        return self._get_pipe(data_pipe=data_pipe, mode=mode)

    def set_array(
        self,
        patch: ArrayLike,
        x: int | slice | None = None,
        y: int | slice | None = None,
        z: int | slice | None = None,
        t: int | slice | None = None,
        c: int | slice | None = None,
        preserve_dimensions: bool = False,
    ) -> None:
        """Set the image data in the zarr array.

        Args:
            patch (ArrayLike): The patch to set.
            x (int | slice | None): The x index or slice.
            y (int | slice | None): The y index or slice.
            z (int | slice | None): The z index or slice.
            t (int | slice | None): The time index or slice.
            c (int | slice | None): The channel index or slice.
            preserve_dimensions (bool): Whether to preserve the dimensions of the data.
        """
        data_pipe = self._build_naive_pipe(
            x=x, y=y, z=z, t=t, c=c, preserve_dimensions=preserve_dimensions
        )
        self._set_pipe(data_pipe=data_pipe, patch=patch)

    def get_array_masked(
        self,
        roi: WorldCooROI,
        t: int | slice | None = None,
        c: int | slice | None = None,
        mask_mode: Literal["bbox", "mask"] = "bbox",
        mode: Literal["numpy"] = "numpy",
        preserve_dimensions: bool = False,
    ) -> ArrayLike:
        """Return the image data from a region of interest (ROI).

        Args:
            roi (WorldCooROI): The region of interest.
            t (int | slice | None): The time index or slice.
            c (int | slice | None): The channel index or slice.
            mask_mode (str): Masking mode
            mode (str): The mode to return the data.
            preserve_dimensions (bool): Whether to preserve the dimensions of the data.
        """
        label_name = roi.infos.get("label_name", None)
        if label_name is None:
            raise ValueError("The label name must be provided in the ROI infos.")

        data_pipe = self._build_roi_pipe(
            roi=roi, t=t, c=c, preserve_dimensions=preserve_dimensions
        )

        if mask_mode == "bbox":
            return self._get_pipe(data_pipe=data_pipe, mode=mode)

        label = self._label_group.get_label(label_name, pixel_size=self.pixel_size)

        mask = label.mask(
            roi,
            t=t,
            mode=mode,
        )
        array = self._get_pipe(data_pipe=data_pipe, mode=mode)
        where_func = np.where if mode == "numpy" else da.where
        return where_func(mask, array, 0)

    def consolidate(self, order: int = 1) -> None:
        """Consolidate the Zarr array."""
        processed_paths = [self]

        todo_image = [
            ImageLike(store=self.group, path=_path)
            for _path in self.metadata.levels_paths
            if _path != self.path
        ]

        while todo_image:
            dist_matrix = np.zeros((len(processed_paths), len(todo_image)))
            for i, image in enumerate(todo_image):
                for j, processed_image in enumerate(processed_paths):
                    dist_matrix[j, i] = np.sqrt(
                        np.sum(
                            [
                                (s1 - s2) ** 2
                                for s1, s2 in zip(
                                    image.shape, processed_image.shape, strict=False
                                )
                            ]
                        )
                    )

            source, target = np.unravel_index(dist_matrix.argmin(), dist_matrix.shape)

            source_image = processed_paths[source]
            target_image = todo_image.pop(target)

            zoom_factors = tuple(
                [
                    s1 / s2
                    for s1, s2 in zip(
                        target_image.shape, source_image.shape, strict=False
                    )
                ]
            )
            partial_zoom = partial(zoom, zoom=zoom_factors, order=order)
            out_image = da.map_blocks(
                partial_zoom,
                source_image.on_disk_dask_array,
                # dtype = source_image.dask_array,
                chunks=target_image.on_disk_array.chunks,
            )
            out_image = out_image.astype(target_image.on_disk_array.dtype)
            # da.to_zarr(out_image, target_image.array, overwrite=True)
            target_image.on_disk_array[...] = out_image.compute()
            # compute the transformation
            processed_paths.append(target_image)



"""A module to handle data transforms for image data."""

from ngio._common_types import ArrayLike
from ngio.pipes._slicer_transforms import SlicerTransform
from ngio.pipes._transforms import Transform


class DataTransformPipe:
    """A class to handle a pipeline of data transforms.

    For example, a pipeline of transforms can be:
        - Select a subset of the data
        - Shuffle the axes of the data
        - Normalize the data

    All these in reverse order will be applied to the data when setting a patch.

    """

    def __init__(self, slicer: SlicerTransform, *data_transforms: Transform):
        """Initialize the DataLoadPipe object.

        Args:
            slicer (SlicerTransform): The first transform to be applied to the
                data MUST be a slicer.
            *data_transforms (Transform): A list of transforms to be
                applied to the data in order.
        """
        self.slicer = slicer
        self.list_of_transforms = data_transforms

    def __repr__(self) -> str:
        """Return the string representation of the object."""
        list_transforms = ", ".join(
            [str(transform) for transform in self.list_of_transforms]
        )
        return f"DataTransformPipe(slicer={self.slicer}, transforms={list_transforms})"

    def get(self, data: ArrayLike) -> ArrayLike:
        """Apply all the transforms to the data and return the result."""
        data = self.slicer.get(data)
        for transform in self.list_of_transforms:
            data = transform.get(data)
        return data

    def set(self, data: ArrayLike, patch: ArrayLike) -> None:
        """Apply all the reverse transforms to the data and return the result."""
        for transform in reversed(self.list_of_transforms):
            patch = transform.set(patch)

        # Write the patch to the data and save it
        self.slicer.set(data, patch)



"""Handler for reading and writing NGFF image metadata."""

from typing import Literal, Protocol

from ngio.io import AccessModeLiteral, Group, StoreOrGroup
from ngio.ngff_meta.fractal_image_meta import ImageLabelMeta
from ngio.ngff_meta.v04.zarr_utils import (
    NgffImageMetaZarrHandlerV04,
)


class NgffImageMetaHandler(Protocol):
    """Handler for NGFF image metadata."""

    def __init__(
        self,
        store: StoreOrGroup,
        meta_mode: Literal["image", "label"],
        cache: bool = False,
        mode: AccessModeLiteral = "a",
    ):
        """Initialize the handler."""
        ...

    @property
    def group(self) -> Group:
        """Return the Zarr group."""
        ...

    @property
    def store(self):
        """Return the Zarr store."""
        ...

    @property
    def zarr_version(self) -> int:
        """Return the Zarr version."""
        ...

    @staticmethod
    def check_version(store: StoreOrGroup = None) -> bool:
        """Check if the version of the metadata is supported."""
        ...

    def load_meta(self) -> ImageLabelMeta:
        """Load the OME-NGFF 0.4 metadata."""
        ...

    def write_meta(self, meta: ImageLabelMeta) -> None:
        """Write the OME-NGFF 0.4 metadata."""
        ...

    def update_cache(self, meta: ImageLabelMeta) -> None:
        """Update the cached metadata."""
        ...

    def clear_cache(self) -> None:
        """Clear the cached metadata."""
        ...


_available_load_ngff_image_meta_handlers = {
    "0.4": NgffImageMetaZarrHandlerV04,
}


def find_ngff_image_meta_handler_version(store: StoreOrGroup) -> str:
    """Find the version of the NGFF image metadata."""
    for version, handler in _available_load_ngff_image_meta_handlers.items():
        if handler.check_version(store=store):
            return version

    supported_versions = ", ".join(_available_load_ngff_image_meta_handlers.keys())
    raise ValueError(
        f"The Zarr store does not contain a supported version. \
            Supported OME-Ngff versions are: {supported_versions}."
    )


def get_ngff_image_meta_handler(
    store: StoreOrGroup,
    meta_mode: Literal["image", "label"],
    version: str | None = None,
    cache: bool = False,
    mode: AccessModeLiteral = "a",
) -> NgffImageMetaHandler:
    """Load the NGFF image metadata handler."""
    if version is None:
        version = find_ngff_image_meta_handler_version(store)

    handler = _available_load_ngff_image_meta_handlers[version]
    return handler(store=store, meta_mode=meta_mode, cache=cache, mode=mode)



class TestTableGroup:
    def test_table_group(self, ome_zarr_image_v04_path):
        from ngio.core.ngff_image import NgffImage

        ngff_image = NgffImage(ome_zarr_image_v04_path)

        ngff_image.table.new(
            name="feat_table",
            table_type="feature_table",
            label_image="region",
            overwrite=True,
        )

        ngff_image.table.new(
            name="roi_table",
            table_type="roi_table",
            overwrite=False,
        )

        ngff_image.table.new(
            name="masking_roi_table",
            table_type="masking_roi_table",
            label_image="region",
            overwrite=False,
        )

        assert ngff_image.table.list() == [
            "feat_table",
            "roi_table",
            "masking_roi_table",
        ]
        assert ngff_image.table.list(table_type="roi_table") == ["roi_table"]
        assert ngff_image.table.list(table_type="feature_table") == ["feat_table"]
        assert ngff_image.table.list(table_type="masking_roi_table") == [
            "masking_roi_table"
        ]



from typing import Protocol

from ngio._common_types import ArrayLike


class Transform(Protocol):
    """A protocol for data transforms to be performed on image data."""

    def get(self, data: ArrayLike) -> ArrayLike:
        """Apply the transform to the data and return the result."""
        ...

    def set(self, data: ArrayLike) -> ArrayLike:
        """Apply the reverse transform to the data and return the result."""
        ...



from importlib.metadata import version
from pathlib import Path
from typing import Literal

import zarr
from packaging.version import Version

zarr_version = version("zarr")
ZARR_PYTHON_V = 2 if Version(zarr_version) < Version("3.0.0a") else 3

# Zarr v3 Imports
# import zarr.store
# from zarr.core.common import AccessModeLiteral, ZarrFormat
# from zarr.store.common import StoreLike

AccessModeLiteral = Literal["r", "r+", "w", "w-", "a"]
ZarrFormat = Literal[2, 3]
StoreLike = str | Path  # This type alias more narrrow than necessary
StoreOrGroup = StoreLike | zarr.Group


class ZarrV3Error(Exception):
    pass


def _pass_through_group(
    group: zarr.Group, mode: AccessModeLiteral, zarr_format: ZarrFormat = 2
) -> zarr.Group:
    if ZARR_PYTHON_V == 2:
        if zarr_format == 3:
            raise ZarrV3Error("Zarr v3 is not supported in when using zarr-python v2.")
        else:
            return group

    else:
        if group.metadata.zarr_format != zarr_format:
            raise ValueError(
                f"Zarr format mismatch. Expected {zarr_format}, "
                "got {store.metadata.zarr_format}."
            )
        else:
            return group

    raise ValueError("This should never be reached.")


def _open_group_v2_v3(
    store: StoreOrGroup, mode: AccessModeLiteral, zarr_format: ZarrFormat = 2
) -> zarr.Group:
    """Wrapper around zarr.open_group with some additional checks.

    Args:
        store (StoreOrGroup): The store (can also be a Path/str) or group to open.
        mode (ReadOrEdirLiteral): The mode to open the group in.
        zarr_format (ZarrFormat): The Zarr format to use.

    Returns:
        zarr.Group: The opened Zarr group.
    """
    if ZARR_PYTHON_V == 3:
        return zarr.open_group(store=store, mode=mode, zarr_format=zarr_format)
    else:
        return zarr.open_group(store=store, mode=mode, zarr_version=zarr_format)


def _is_group_readonly(group: zarr.Group) -> bool:
    if ZARR_PYTHON_V == 3:
        return group.store_path.store.mode.readonly

    else:
        return not group.store.is_writeable()



"""Next Generation file format IO."""

from importlib.metadata import PackageNotFoundError, version

from ngio.core import Image, Label, NgffImage

__all__ = ["Image", "Label", "NgffImage"]

try:
    __version__ = version("ngio")
except PackageNotFoundError:  # pragma: no cover
    __version__ = "uninstalled"
__author__ = "Lorenzo Cerrone"
__email__ = "lorenzo.cerrone@uzh.ch"



class TestLabel:
    def test_label_image(self, ome_zarr_image_v04_path):
        import dask.array as da
        import numpy as np

        from ngio.core.ngff_image import NgffImage

        ngff_image = NgffImage(ome_zarr_image_v04_path)
        image_handler = ngff_image.get_image(path="0")
        label_handler = ngff_image.label.derive(name="label")

        assert ngff_image.label.list() == ["label"]
        assert "c" not in label_handler.axes_names
        assert label_handler.shape == image_handler.shape[1:]

        shape = label_handler.shape
        assert label_handler.get_array(mode="dask").shape == shape

        label_handler.set_array(patch=da.ones((10, 256, 256), dtype=np.uint16))
        assert label_handler.get_array(t=0, z=0, x=0, y=0) == 1

        label_handler.consolidate()

        label_handler_1 = ngff_image.label.get_label(name="label")
        assert label_handler_1.get_array(t=0, z=0, x=0, y=0) == 1
